package element;
import java.util.HashMap;

public class SymbolTable
{
	
	/**
	 * Instancia nica (Singleton) de SymbolTable.
	 */
	private static SymbolTable instance = null;
	
	/**
	 * Tamao por defecto con el que comienza el array de la tabla de smbolos.
	 */
	private static int DEFAULT_INITIAL_SIZE = 50;
	
	/**
	 * Relaciona cada entero (ID de smbolo) con el nombre del smbolo correspondiente.
	 */
	private String[] symbolNames = new String[DEFAULT_INITIAL_SIZE];

	/**
	 * Almacena los smbolos en s, relacionando cada ID con el smbolo correspondiente.
	 */
	private Symbol[] symbols = new Symbol[DEFAULT_INITIAL_SIZE];
	
	/**
	 * Tabla hash que vincula nombres a smbolos.
	 */
	private HashMap symbolsForNames = new HashMap();
	
	/*Asercin permanente: symbolNames.length == symbols.length*/
	
	/**
	 * ltima ID de smbolo que ha sido asignada a algn smbolo. Las ID's se irn
	 * asignando por orden y el array symbolNames crecer cuando sea necesario.
	 */
	private int lastIdAssigned = -1;
	
	/**
	 * Mtodo de acceso al Singleton.
	 * @return La instancia nica de SymbolTable que existe en el sistema.
	 */
	public static SymbolTable instance()
	{
		if ( instance == null )
			instance = new SymbolTable();
		return instance;
	}
	
	/**
	 * Borra la tabla de smbolos.
	 *
	 */
	public static void reset()
	{
		instance = new SymbolTable();
	}
	
	/**
	 * Devuelve la ltima ID que ha sido asignada.
	 *
	 */
	public int getLastId()
	{
		return lastIdAssigned;
	}
	
	/**
	 * Constructor privado del Singleton.
	 *
	 */
	private SymbolTable()
	{
		;
	}
	
	/** Nos da el smbolo conocido de nombre dado, si existe, y si no lo crea (como ground).
	* Ntese que no se puede crear un smbolo ground y otro no-ground
	 * con el mismo nombre.
	 * @param name Nombre del smbolo deseado.
	 * @return Instancia de Symbol correspondiente al nombre dado.
	 */
	public Symbol symbolFor ( String name )
	{
		//ver si el smbolo ya existe:
		Symbol existing = (Symbol) symbolsForNames.get(name);
		if ( existing != null ) return existing;
		//crear un nuevo smbolo:
		int id = (++lastIdAssigned);
		if ( id < symbolNames.length ) //or symbols.length, son la misma
		{
			symbolNames[id] = name;
			symbols[id] = new Symbol(id,true);
			symbolsForNames.put(name,symbols[id]);
		}
		else
		{
			String[] newSymbolNames = new String[(int)(1.5*symbolNames.length)];
			for ( int i = 0 ; i < symbolNames.length ; i++ )
			{
				newSymbolNames[i] = symbolNames[i];
			}
			symbolNames = newSymbolNames;
			//assert (id < symbolNames.length);
			symbolNames[id] = name;
			Symbol[] newSymbols = new Symbol[(int)(1.5*symbols.length)];
			for ( int i = 0 ; i < symbols.length ; i++ )
			{
				newSymbols[i] = symbols[i];
			}
			symbols = newSymbols;
			//assert (id < symbols.length);
			symbols[id] = new Symbol(id,true);
			symbolsForNames.put(name,symbols[id]);
		}
		return symbols[id];		
	}
	
	/**
	 * Devuelve el nombre de smbolo asignado a la ID dada.
	 * @param id Identificador del smbolo cuyo nombre queremos recuperar.
	 * @return Nombre del smbolo cuya ID es id, o bien null si dicho
	 * nmero ID est sin asignar.
	 */
	public String getName ( int id )
	{
		try
		{
			return symbolNames[id];
		}
		catch ( Exception exc )
		{
			return null;
		}
	}
	
	/**
	 * Devuelve el smbolo de ID dada.
	 * @param id Identificador del smbolo que queremos recuperar.
	 * @return El smbolo cuya ID es id, o bien null si dicho nmero
	 * ID est sin asignar.
	 */
	public Symbol getSymbol ( int id )
	{
		try
		{
			return symbols[id];
		}
		catch ( Exception exc )
		{
			return null;
		}
	}
	
	/**
	 * Devuelve el nombre del smbolo dado.
	 * @param s Smbolo del que se quiere saber el nombre.
	 * @return Nombre del smbolo. Este mtodo no tirar excepciones si el smbolo no es
	 * null, y si todos los smbolos han sido creados mediante SymbolTable::SymbolFor(String).
	 */
	public String getName ( Symbol s )
	{
		return getName(s.getID());
	}
		
	
	
}