package element.tag;

import java.util.HashMap;

import element.HashIndexable;
import element.SimpleElement;
import element.Symbol;
import element.SymbolTable;

/**
 * This class represents one of the nodes of a tree-adjoining grammar tree.
 * It contains a reference to the tree and a node index.
 * @author Carlos
 *
 */
public class TAGTreeNode extends SimpleElement implements HashIndexable
{
	
	private TAGTree tree;
	private int nodeIndex;
	private boolean ground;
	
	private boolean nullAdjunction;
	private boolean substitutionNode;
	
	public TAGTreeNode ( TAGTree tree , int nodeIndex )
	{
		this.tree = tree;
		this.nodeIndex = nodeIndex;
		this.ground = true;
	}
	
	public TAGTreeNode ( TAGTree tree , int nodeIndex , boolean nullAdjunction )
	{
		this(tree,nodeIndex);
		this.nullAdjunction = nullAdjunction;
	}
	
	public TAGTreeNode ( TAGTree tree , int nodeIndex , boolean nullAdjunction , boolean substitutionNode )
	{
		this(tree,nodeIndex,nullAdjunction);
		this.substitutionNode = substitutionNode;
	}

	public String toString()
	{
		if ( ground )
		{
			if ( getAssociatedSymbol() != null )
				return getAssociatedSymbol().toString() + nodeIndex + "@" + tree.getName();
			else
				return "null" + "@" + tree.getName();
		}
		else
		{
			return super.toString() + " ng ";
		}
	}
	
    public Object getHashKey()
    {
        return this;
    }
	
	/**
	 * Builds a NON-GROUND TAG tree node.
	 *
	 */
	public TAGTreeNode ( )
	{
		this.ground = false;
	}
	
	public boolean isGround()
	{
		return ground;
	}
	
	public Symbol getAssociatedSymbol()
	{
		return tree.symbolForNode(nodeIndex);
	}
	
	public TAGTree getAssociatedTree()
	{
		return tree;
	}
	
	/**
	 * Returns whether the given tree can be adjoined at this node.
	 * Takes into account the null adjunction (NA) constraint and the symbol
	 * in the given tree's root node.
	 * @param tree A TAG tree.
	 * @return true if the given tree can be adjoined at this node, false otherwise.
	 */
	public boolean adjoinable ( TAGTree tree )
	{
		if ( nullAdjunction ) return false;
		if ( tree.isInitial() ) return false;
		if ( !tree.symbolForRoot().equals(this.getAssociatedSymbol())) return false;
		return true;
	}
	
	/**
	 * Returns whether the given tree can be substituted at this node.
	 * Takes into account the substitution node status of this node, whether
	 * the given tree is initial, and the symbol in its root node.
	 * @param tree A TAG tree.
	 * @return true if the given tree can be substituted at this node, false otherwise.
	 */
	public boolean substitutable ( TAGTree tree )
	{
		if ( !substitutionNode ) return false;
		if ( !tree.isInitial() ) return false;
		if ( !tree.symbolForRoot().equals(this.getAssociatedSymbol())) return false;
		return true;
	}
	
	public boolean isSubstitutionNode()
	{
		return substitutionNode;
	}
	
	private static HashMap nonGroundInstances = new HashMap();
	
	/**
	 * Obtains a nonground TAG tree node represented by a given string.
	 * @param s String representing the nonground TAG tree node.
	 * @return Nonground TAG tree node.
	 */
	public static TAGTreeNode nonGroundFromString ( String s )
	{
		TAGTreeNode node = (TAGTreeNode) nonGroundInstances.get(s);
		if ( node == null )
		{
			node = new TAGTreeNode();
			nonGroundInstances.put(s,node);
		}
		return node;
	}
	
}
