package engine;

import java.util.List;

import schema.step.DeductiveStep;
import schema.item.Items;
import schema.item.ItemHandler;
import schema.step.StepHandler;
import java.util.List;

import agenda.Agenda;


/**
 * This class overrides the process() method of the standard deductive engine to provide a version of the deductive engine with support
 * for regional error correction.
 * This machine works by increasing an error counter or decreasing a retract counter when the agenda is empty but no final (goal) items have
 * been found. Therefore, the schemata to be executed in this machine are supposed to have (1) at least a goal specification, (2) error-correcting
 * deductive steps which use the error and retract counter as conditions, so that the necessary corrections can be undertaken when goal items
 * are not found and counters are modified.
 * 
 * @author Carlos Gmez
 * Created at mortadelo, 21/11/2006 13:02:13
 *
 */
public class RegionalErrorCorrectingDeductiveEngine extends StandardDeductiveEngine
{
	

	public RegionalErrorCorrectingDeductiveEngine ( Agenda agenda , List deductiveSteps , List goalConditions , ItemHandler itemHandler , StepHandler stepHandler )
	{
		super ( agenda , deductiveSteps , goalConditions , itemHandler , stepHandler );
	}
	
	
	/**
	 * Returns whether the given list contains final items according to the goal conditions associated to the schema
	 * this deductive engine was instantiated with.
	 */
	private boolean finalItemsFound( List lItems )
	{
		
		boolean found = false;
		for ( int i = 0 ; i < lItems.size() ; i++ )
		{
			Object[] item = (Object[]) lItems.get(i);
			if ( goalConditions == null || goalConditions.size() == 0 ) { found = true ; break; }
			else
			{
				for ( int j = 0 ; j < goalConditions.size() ; j++ )
				{
					List result = ((DeductiveStep)goalConditions.get(j)).applyTo(item);
					if ( result != null && result.size() > 0 )
					{
						System.out.println("A final item: " + Items.toString(item));
						found = true;
						break; //esto slo rompe el for externo
					}
				}
				if ( found ) break;
			}
		}
		return found;
		
	}
	
	
	private int retractCount = 0;
	private int errorCount = 0;
	private int rightmostPositionScanned = 0;
	private int detectionCount = 0;
	
	public int getRetractCount()
	{
		return retractCount;
	}
	
	public int getErrorCount()
	{
		return errorCount;
	}
	
	public int getDetectionCount()
	{
		return detectionCount;
	}
	
	
	/*
	public void setRightmostPositionScanned ( int pos )
	{
		this.rightmostPositionScanned = pos;
	}
	*/
	
	
	
	private int rightmostNextToExamine = 0;
	
	public int getRightmostPositionScanned ( )
	{
		List items = itemHandler.getAll();
		for ( ; rightmostNextToExamine < items.size() ; rightmostNextToExamine++ )
		{
			Object[] item = (Object[]) items.get(rightmostNextToExamine);
			if ( item[0] instanceof element.SpecialElement && ((element.SpecialElement)item[0]).toString().equals("Parse") )
			{
				if ( item[3] instanceof element.StringPosition )
				{
					int newVal = ((element.StringPosition)item[3]).getValue();
					if ( rightmostPositionScanned < newVal ) rightmostPositionScanned = newVal;
				}
			}
		}
		System.out.println("Rightmost pos " + rightmostPositionScanned);
		System.out.println("I analyzed " + items.size() + " items");
		return rightmostPositionScanned;
	}
	
	
	public List preprocess()
	{	
		return super.process();
	}
	
	
	public static int MAX_ERRORS = 3;
	
	public List process()
	{
	
		retractCount = 0;
		errorCount = 0;
		rightmostPositionScanned = 0;
		rightmostNextToExamine = 0;
		detectionCount = 0;
		
		boolean done = false;
		
		for (;;)
		{
			
			System.out.println("Going to launch standard processing. [e="+errorCount+", r="+retractCount+", d="+detectionCount+"]");
			
			//execute the standard deductive engine until agenda is empty
			List lItems = super.process();
			
			System.out.println("Standard processing done. [e="+errorCount+", r="+retractCount+", d="+detectionCount+"]");
			
			//check whether final items have been found
			if ( finalItemsFound(lItems) ) return lItems;
			else
			{
				
				System.out.println("Final items not found.");
				
				//Re-insert items into agenda for error parsing
				for ( int i = 0 ; i < lItems.size() ; i++ )
					agenda.add(lItems.get(i));
				
				if ( retractCount > 0 )
				{
					retractCount--;
					
					//[added 2006-11-19]
					//?? -> this way it can go right if necessary.
					int newDetection = getRightmostPositionScanned();
					if ( newDetection > detectionCount )
					{
						detectionCount = newDetection;
						retractCount = detectionCount;
					}
				
				}
				else
				{
					detectionCount = getRightmostPositionScanned();
					retractCount = detectionCount;
					errorCount++;
					
					if ( errorCount > MAX_ERRORS )
					{
						System.out.println("No parsers found for max error count " + MAX_ERRORS + ". Sorry.");
						return lItems;
					}
					
				}
				
				System.out.println("Retr " + retractCount + ", err " + errorCount);
				
			}
			
		}
		
		
	}
	
	public Object getProperty ( String key )
	{
		if ( key.equals("retractCount") )
			return new Integer(retractCount);
		if ( key.equals("errorCount") )
			return new Integer(errorCount);
		if ( key.equals("detectionCount") )
			return new Integer(detectionCount);
		return null;
	}
	

	
	
}