package engine;

import java.util.List;

import schema.step.DeductiveStep;
import schema.item.ItemHandler;
import schema.step.StepHandler;
import java.util.List;

import agenda.Agenda;

/**
 * Standard deductive engine, as the one described by Shieber, with the agenda implemented as a queue.
 * 
 * @author Carlos Gmez
 * Created at mortadelo, 21/11/2006 12:48:17
 */
public class StandardDeductiveEngine implements DeductiveEngine
{
	
	protected Agenda agenda;
	protected List deductiveSteps;
	protected List goalConditions;
	protected ItemHandler itemHandler;
	protected StepHandler stepHandler;
	
	public void setItemHandler ( ItemHandler ih )
	{
		itemHandler = ih;
	}
	
	public StandardDeductiveEngine ( Agenda agenda , List deductiveSteps , List goalConditions , ItemHandler itemHandler , StepHandler stepHandler )
	{
		this.agenda = agenda;
		this.deductiveSteps = deductiveSteps;
		this.goalConditions = goalConditions;
		this.itemHandler = itemHandler;
		this.stepHandler = stepHandler;
		EngineManager.register(this);
	}
	
	
	/**
	 * Generates all the items that can be deducted from the items in the agenda.
	 * In this concrete class, process() is the same as preprocess().
	 * The standard deductive engine just triggers all possible deductions with the items in the agenda, without any further consideration.
	 * This is valid both for preprocessing when only the grammar is known (and steps requiring no knowledge of the string, such as
	 * Earley Initter, are executed) and for executing the actual algorithm.
	 * @return List of all the items in the ItemHandler.
	 */
	public List preprocess ()
	{
		return process ();
	}
	
	
	private void processEmptyAntecedentSteps()
	{
		
		//Aplicar pasos de antecedente vaco y quitarlos de la lista de pasos aplicables.
		int sz = deductiveSteps.size();
		Object[] emptyItem = new Object[0];
		for ( int i = sz-1 ; i >= 0 ; i-- )
		{
			List l = ((DeductiveStep)deductiveSteps.get(i)).applyTo(emptyItem);
			if ( l!=null )
			{
				for ( int j = 0 ; j < l.size() ; j++ )
				{
						Object[] newItem = (Object[]) l.get(j);
						if ( !itemHandler.exists(newItem) )
						{
							itemHandler.addItem(newItem);
							agenda.add(newItem);
						}
				}
				if ( l.size() > 0 )
				{
					if ( ((DeductiveStep)deductiveSteps.get(i)).hasEmptyAntecedent() )
						deductiveSteps.remove(i);
				}
			}
		}
		
	}
	
	
	/**
	 * Generates all the items that can be deducted from the items in the agenda.
	 * In this concrete class, process() is the same as preprocess().
	 * The standard deductive engine just triggers all possible deductions with the items in the agenda, without any further consideration.
	 * This is valid both for preprocessing when only the grammar is known (and steps requiring no knowledge of the string, such as
	 * Earley Initter, are executed) and for executing the actual algorithm.
	 * @return List of all the items in the ItemHandler.
	 */
	public List process()
	{
		
		//Aplicar pasos de antecedente vaco y quitarlos de la lista de pasos aplicables.
		processEmptyAntecedentSteps();
		
		int sz = 0;
		//aplicar pasos con antecedente, usando agenda
		while ( !agenda.isEmpty() )
		{
			Object[] curItem = (Object[]) agenda.removeFirst();
			
			List pasosDeductivosAplicables = stepHandler.getByItem(curItem);
			
			sz = pasosDeductivosAplicables.size();
			for ( int i = 0 ; i < sz ; i++ )
			{
				List l = ((DeductiveStep)pasosDeductivosAplicables.get(i)).applyTo(curItem);
				if ( l != null )
				{
					int sz2 = l.size();
					for ( int j = 0 ; j < sz2 ; j++ )
					{
						Object[] newItem = (Object[]) l.get(j);
						if ( !itemHandler.exists(newItem) )
						{
							itemHandler.addItem(newItem);
							agenda.add(newItem);
						}
					}
				}
			}
		}
		
		List lItems = (itemHandler.getAll());
				
		return lItems;
		
	}
	
	
	
	
	public Object getProperty ( String key )
	{
		return null;
	}
	

	
	
	
	
	
	
}