/*
 * Created on 09-ago-2004 11:50:15
 *
 */
package generator;


import element.CompositeElement;
import element.ExpressionElement;
import element.SequenceElement;

import java.util.*;
import element.SimpleElement;

/**
 * <p>Esta clase genera el cdigo de los mtodos applyToI correspondiente a llevar a 
 * cabo el "pattern-matching" entre el item del antecedente que maneja ese mtodo concreto
 * y un item que se nos pasa como parmetro en una variable de nombre objectName (el
 * llamado "trigger item").</p>
 * <p>Este "pattern-matching" incluye comprobar si el item dado es el item del antecedente
 * o un caso particular suyo, y asignar las variables "no-ground" que haya en el item
 * del antecedente a sus valores concretos del item dado.</p>
 * <p>Created 09-ago-2004 11:50:15</p>
 * @author carlos
 */
public class ApplyTriggerMatchVisitor extends CodeGeneratingVisitor 
{

	/**
	 * En este buffer se almacena el cdigo que devuelve el paso deductivo a su
	 * estado inicial si descubrimos que no hay matching cuando ya se haban llevado
	 * a cabo algunas asignaciones.
	 */
	private StringBuffer recoveryCode = new StringBuffer ( "" );
	
	/**
	 * A medida que vamos encontrando atributos no-ground que pueden hacer matching,
	 * les vamos cambiando el estado en el generador a ground. Tenemos que almacenar
	 * qu atributos hemos cambiado de estado porque ese cambio no valdr para los
	 * restantes mtodos applyTo que no sean el que estamos generando ahora, teniendo
	 * que deshacerse la operacin de "groundificacin" al final de la generacin
	 * del cdigo del mtodo actual. Esto lo hace la funcin undoChanges().
	 */
	private List groundedAttrs = new ArrayList();
    
    /**
     * A esta cadena se le dar un valor cuando este visitante visite un elemento
     * de tipo Sequence. Este valor ser el nombre de la variable que contiene
     * el nmero de items que han emparejado con el Sequence. Tal nombre es necesario
     * almacenarlo para que leerlo cuando estemos visitando el CompositeElement que
     * contiene el Sequence, pues lo utilizaremos para aadirlo al ndice del array.
     */
    private String indexIncrement = null;
	
	/**
	 * Obtiene el buffer donde se almacena el cdigo de recuperacin.
	 * Este cdigo tiene que ser utilizado por el generador de cdigo y otros
	 * visitantes ms adelante.
	 * @return Buffer con cdigo de recuperacin del estado del paso deductivo para 
	 * contrarrestar lo que hace el cdigo generado por este visitante, en caso
	 * de que no haya matching.
	 */
	public StringBuffer getRecoveryCodeBuffer()
	{
	    return recoveryCode;
	}
	
	/**
	 * @param objectName Nombre del item contra el que hacemos pattern-matching.
	 * @param toAppendTo Buffer donde insertaremos el cdigo generado.
	 * @param cg Generador de cdigo que contiene la informacin sobre atributos.
	 */
	public ApplyTriggerMatchVisitor ( String objectName, StringBuffer toAppendTo, DeductiveStepCodeGenerator cg ) 
	{
		super ( objectName, toAppendTo, cg );
	}

	/**
	 * Genera el cdigo para un elemento simple.
	 */
	public void visitSimpleElement(SimpleElement elt) 
	{
	    Attribute attr = generator.attributeFor ( elt );

	    //assertion: {attr != null}.
	    //todos los elementos hoja del antecedente dan lugar a atributos, y los
	    //elementos visitados deberan pertenecer a un item del antecedente.
	    
	    boolean grounded = generator.isGround(attr);
	    
	    if ( attr == null )
	    {
	    	//atributo ground desde el principio.
	    	toAppendTo.append( "if ( !" + objectName + ".equals(" + elt.getExpressionCode() + ") )\n{\n"+recoveryCode.toString()+"return result;\n}\n");
	    }
	    else if ( grounded )
	    {
	        //atributo ya asignado, se ha de chequear su valor.
	        toAppendTo.append( "if ( !" + objectName + ".equals(" + attr.getName() + ") )\n{\n"+recoveryCode.toString()+"return result;\n}\n" );
	    }
	    else
	    {
	        //atributo no asignado, se le ha de asignar el valor correspondiente en el item dado,
	        //con cuidado porque las clases podran no coincidir, en cuyo caso debemos
	        //generar el "recovery code" para salir del mtodo sin un match.
	        
	        //declaracin de variable old
	    	
	    	toAppendTo.append(attr.getClassName() + " " + "old_" + attr.getName() + " = " + attr.getName() + ";\n" );
	        
	        //agregar asignacin inversa al cdigo de recuperacin
	        recoveryCode.insert(0, attr.getName() + " = " + "old_" + attr.getName() + ";\n" );

	        //chequeo de clase
	        //optimized version (no speculative cast)
	        toAppendTo.append ( "if ( " + objectName + " instanceof " + attr.getClassName() + " )\n" );
	        toAppendTo.append ( "{\n" );
	        toAppendTo.append ( attr.getName() + " = " + "( " + attr.getClassName() + " )" + " " + objectName + ";\n" );
	        toAppendTo.append ( "} //class check passed\n" );
	        toAppendTo.append ( "else\n" );
	        toAppendTo.append ( "{\n" );
	        toAppendTo.append ( recoveryCode.toString() );
	        toAppendTo.append ( "return result;\n" );
	        toAppendTo.append ( "} //class check failed\n" );
	        
	        	        
	        generator.setGround ( attr , true );
	        groundedAttrs.add(attr);
	        
	    }
	    
	}

    /**Flag que slo se activa en la primera llamada al visitante, que ser la ms externa, despus estar a false
     * para que el resto de llamadas recursivas.*/
    private boolean firstTimeFlag = true;
	
    
	/**
	 * En este caso no se hace nada para los elementos compuestos salvo pasar a los hijos,
	 * pues las asignaciones en principio sern siempre de objetos elementales.
	 * Esto es lo mismo que para ApplyNonTriggerAssignVisitor.
	 */
    public void visitCompositeElement ( CompositeElement ce /*, String increments*/ )
    {
		String oldObjectName = this.objectName;
		indexIncrement = null;
		String increments = "";
		
		boolean externalFlag = firstTimeFlag;
		firstTimeFlag = false;
		
		//hallamos el nmero de hijos que no son zero or more's
		int nonSequenceChildren = 0;
		for ( int i = 0 ; i < ce.numChildren() ; i++ )
		{
			if ( !(ce.getChild(i) instanceof SequenceElement ))
				nonSequenceChildren++;
		}
		int nonSequenceChildrenProcessed = 0; //processed since last setting of variable increments
		int nonSequenceChildrenLeft = nonSequenceChildren; //left to process, in general
		for ( int i = 0 ; i < ce.numChildren() ; i++ )
		{	
			if ( indexIncrement != null )
			{
				/**
				 * Si entramos en este if, es que el ltimo hijo procesado era un
				 * Sequence (multimatcher). En este caso, no debemos incrementar
				 * el ndice en 1, pues no sabemos en tiempo de generacin de cdigo
				 * si emparejaremos con un elemento, con cero o con ms de uno.
				 * El nombre de la variable que contendr el nmero de elementos con
				 * que emparejamos lo guarda el mtodo visitSequenceElement de 
				 * este mismo visitante en la variable indexIncrement. As pues,
				 * tenemos que sumar ese nombre a nuestro ndice en el cdigo generado.
				 */
				//increments+=("+"+indexIncrement);
				increments = "+"+indexIncrement;
				//indexIncrement = null; //later!
				nonSequenceChildrenProcessed = 0;
				/**
				 * Aqu necesitamos un chequeo intermedio de longitud.
				 */
				String tmpname = (externalFlag)? (oldObjectName) : ("((Object[])"+oldObjectName+")"); 
				if ( !externalFlag ) 
				{
					toAppendTo.append("if ( " + oldObjectName + " instanceof Object[] )\n");
					toAppendTo.append("{\n");
				}
				if ( nonSequenceChildrenLeft == ce.numChildren()-i /*total children left*/ )
					toAppendTo.append ( "if ( " + tmpname + ".length != " + (nonSequenceChildrenLeft) + "+" + indexIncrement + " )\n" );
		        else //no quedan ms multimatchers por matchear, la longitud ha de ser exacta
		        	toAppendTo.append ( "if ( " + tmpname + ".length < " + (nonSequenceChildrenLeft) + "+" + indexIncrement + " )\n" );
				toAppendTo.append ( "{\n" );
		        toAppendTo.append ( recoveryCode.toString() );
		        toAppendTo.append ( "return result;\n" );
		        toAppendTo.append ( "}\n" ); 
		        if ( !externalFlag ) 
		        {
		        	toAppendTo.append("}\n");
		        	toAppendTo.append("else\n");
			        toAppendTo.append("{\n");
			        toAppendTo.append ( recoveryCode.toString() );
			        toAppendTo.append ( "return result;\n" );
			        toAppendTo.append ( "}\n" );
		        }
		        indexIncrement=null;
			}
			if ( !externalFlag )
				this.objectName = "((Object[])" + oldObjectName + ")" + "["+nonSequenceChildrenProcessed+increments+"]";
			else
				this.objectName = oldObjectName + "["+nonSequenceChildrenProcessed+increments+"]";
			ce.getChild(i).accept(this);
			if ( !(ce.getChild(i) instanceof SequenceElement ))
			{
				nonSequenceChildrenProcessed++;
				nonSequenceChildrenLeft--;
			}
		}
		if ( indexIncrement != null )
		{
			/*
			 * Necesitamos un chequeo final de longitud, porque podemos haber hecho
			 * pattern-matching con todo pero que todava nos queden cosas en el item.
			 */
			increments="+"+indexIncrement;
			String tmpname = (externalFlag)? (oldObjectName) : ("((Object[])"+oldObjectName+")"); 
			if ( !externalFlag ) 
			{
				toAppendTo.append ( "if ( " + oldObjectName + " instanceof Object[] )\n" );
				toAppendTo.append ( "{\n" );				
			}
			//{nonSequenceChildrenLeft==0}
			toAppendTo.append ( "if ( " + tmpname + ".length != " + (nonSequenceChildrenLeft) + "+" + indexIncrement+ " )\n" );
	        toAppendTo.append ( "{\n" );
	        toAppendTo.append ( recoveryCode.toString() );
	        toAppendTo.append ( "return result;\n" );
	        toAppendTo.append ( "}\n" ); 
	        if ( !externalFlag ) 
	        {
				toAppendTo.append ( "} //class is Object[]\n" );
		        toAppendTo.append ( "else\n" );
		        toAppendTo.append ( "{\n" );
		        toAppendTo.append ( recoveryCode.toString() );
		        toAppendTo.append ( "return result;\n" );
		        toAppendTo.append ( "} //class is not Object[]\n" ); 
	        }
			
		
			indexIncrement = null; //Si el ltimo elemento visitado ha sido un Sequence,
			//no queremos que esto afecte al llamador externo.
		
		}
		this.objectName = oldObjectName; //dejamos el visitante como estaba.
    }
	
	
	 /**
     * En el caso de un elemento de tipo Sequence, este visitador hace bsicamente
     * lo mismo que el ApplyNonTriggerAssignVisitor, dado que la posibilidad de
     * que no haya matching realmente no existe (puede haber matching de cero
     * elementos o de ms; pero lo habr).
     * tenemos que generar cdigo
     * para ir aadiendo al atributo asociado a ese elemento, que ser una lista,
     * todos los elementos simples de la clase correspondiente a ese Sequence
     * que nos encontremos a partir de objectName (el propio objectName y los
     * siguientes, obtenidos modificando el ndice de array de objectName).
     * @param zome
     */
    public void visitSequenceElement ( SequenceElement zome )
    {
    	Attribute attr = generator.attributeFor ( zome );
    	//{attr not null y attr.getElementClass() debera ser alguna subclase de List}
    	if ( !generator.isGround(attr) )
    	{
    		
	        //agregar asignacin inversa al cdigo de recuperacin.
    		//En este caso no necesitamos variable old, sabemos que si el Sequence
    		//era no ground la lista iba a estar vaca.
		    recoveryCode.insert(0, attr.getName() + ".clear();\n");
    		
	        //Tomar el ndice de objectName.
	        //Aqu partimos la cadena objectName para separar en partes
	        //el ndice del array, la parte anterior, la parte posterior y el nombre
	        //del array. Por ejemplo, si tenemos
	        //miItem[k][6][i+1] => hacemos:
	        //beforeIndex = miItem[k][6][
	        //index = i+1
	        //afterIndex = ]
	        //arrayName = miItem[k][6] (el arrayName se usa para ver la longitud).
	        int lastOpenBracket = objectName.lastIndexOf("[");
	        int closedBracket = objectName.indexOf("]",lastOpenBracket);
	        String beforeIndex = objectName.substring(0,lastOpenBracket+1);
	        String index = objectName.substring(lastOpenBracket+1,closedBracket);
	        String afterIndex = objectName.substring(closedBracket);
	        String arrayName = beforeIndex.substring(0,beforeIndex.length()-1) + afterIndex.substring(1);
	        
	        //obtenemos el nombre para el ndice de bucle de este multimatcher.
	        String loopIndexName = attr.getName() + "_index";
	        
	        //Generamos el cdigo de un bucle que ir recorriendo elementos y aadindolos
	        //a la lista del multimatcher mientras sean de la clase asociada a ste.
	        //Ntese que en este caso no necesitamos estructuras try-catch
	        //ni cdigo de recuperacin, porque el matching de un Sequence no puede
	        //fallar: en el caso de que nos encontremos elementos cuya clase no
	        //coincide con la esperada por el Sequence, simplemente ste habr
	        //emparejado con cero elementos; pero empareja correctamente.
	        
	        //REMOVED this line @ 05.12.22:
	        //toAppendTo.append ( attr.getName() + ".clear();\n" );
	        //We don't need it. As an applyTo precondition, the array lists are
	        //clear. They are cleared at the recovery code which is always execced
	        //before returning an item list (even an empty one)
	        
	        toAppendTo.append ( "int " + loopIndexName + " = " + index + ";\n");
	        toAppendTo.append ( "while ( " + loopIndexName + " < " + arrayName + ".length && " + beforeIndex + loopIndexName + afterIndex + " instanceof " + zome.getAssociatedClass().getName() + " )\n" );
	        toAppendTo.append ( "{\n" );
	        toAppendTo.append ( attr.getName() + ".add(" + beforeIndex + loopIndexName + afterIndex + ");\n" );
	        toAppendTo.append ( loopIndexName + "++;\n");
	        toAppendTo.append ( "}\n");
	        
	        //Fijamos el valor de indexIncrement, utilizado por el mtodo visitCompositeElement
	        //que llama a ste.
	        indexIncrement = loopIndexName;
	        
	        //Indicamos en el generador de cdigo que la lista ya es ground.
	        generator.setGround ( attr , true );
	        groundedAttrs.add(attr);
    		
    	}
    	else
    	{
    		System.out.println("[ERROR] Modify this.");
    	}
    	
    }

	
	public void visitExpressionElement ( ExpressionElement ee )
	{
	    boolean ground = generator.isGround(ee);
	    if ( !ground )
	    {
	        /*
	         * esto sucede seguramente porque no podemos considerar los items del
	         * antecedente en este orden, antes de utilizar ste deberamos haber
	         * usado algn otro que nos concretara los miembros de la expresin.
	         * Por lo tanto, si llegamos en este orden, simplemente nos recuperamos
	         * y no aadimos nada a result.
	         */
	    	//Este planteamiento se podra mejorar con simplificaciones.
	        toAppendTo.append ( recoveryCode.toString() );
	        toAppendTo.append ( "return result;\n" );
	    }
	    else
	    {
	        //atributo ya asignado, se ha de chequear su valor.
	        toAppendTo.append( "if ( !" + objectName + ".equals(" );
	        toAppendTo.append(generator.generateExpressionCode(ee));
	        toAppendTo.append ( ") )\n{\n"+recoveryCode.toString()+"return result;\n}\n" );
	    }
	}
	
	
	/**
	 * Esta funcin hace que los atributos que habamos fijado a ground en el generador
	 * de cdigo vuelvan a no-ground, para poder generar otros mtodos.
	 * Tiene que ser llamada desde fuera en el momento en que se acabe de generar
	 * el mtodo applyTo que incluye el cdigo generado por este visitante.
	 */
	public void undoChanges()
	{
	    for ( int i = 0 ; i < groundedAttrs.size() ; i++ )
	    {
	        generator.setGround ( (Attribute)groundedAttrs.get(i) , false );
	    }
	}
	
	

}
