package generator;

import parser.sparser.tree.*; //partimos del parse tree dado por el schema parser
import util.FileCopy;
import java.util.*;
import java.io.*;

import agenda.Agenda;

/**
 * @author carlos
 *
 * sta clase se encarga de generar el cdigo en que se traduce un esquema
 * de anlisis sintctico (parsing schema). Esto se traduce en:
 * - Clases asociadas a cada paso deductivo del esquema.
 * - Clases para la indexacin de items y pasos deductivos.
 * Para llevar a cabo la generacin de cdigo, CodeGenerator llama a otras clases
 * (ItemHandlerCodeGenerator, StepHandlerCodeGenerator, DeductiveStepCodeGenerator) que
 * se encargan de generar las diferentes clases. 
 */
public class CodeGenerator 
{
	
	private SchemaNode sn; //nodo raz del arbol que representa el esquema de anlisis sintctico.
	
	private final boolean VERBOSE_OUTPUT; //si se deben sacar salidas por la lnea de comandos.
	
	private String outputDirectory = "generated"+File.separator; //directorio de salida por defecto.
		
	/**
	 * Construye un objeto para generar el cdigo asociado a un esquema de anlisis
	 * sintctico dado.
	 * @param sn Nodo raz del rbol que representa el esquema de anlisis sintctico
	 * cuyo cdigo podr ser generado por el CodeGenerator que construimos.
	 */
	public CodeGenerator ( SchemaNode sn )
	{
		this.sn = sn;
		VERBOSE_OUTPUT = true;
	}
	
	/**
	 * Construye un objeto para generar el cdigo asociado a un esquema de anlisis
	 * sintctico dado.
	 * @param sn Nodo raz del rbol que representa el esquema de anlisis sintctico.
	 * @param verbose Si se deben sacar salidas por la lnea de comandos.
	 * @param outputDirectory Directorio donde se debe escribir el cdigo.
	 * cuyo cdigo podr ser generado por el CodeGenerator que construimos.
	 */
	public CodeGenerator ( SchemaNode sn , boolean verbose , String outputDirectory )
	{
		this.sn=sn;
		this.VERBOSE_OUTPUT = verbose;
		this.outputDirectory = outputDirectory;
	}
	
	/**
	 * Genera el cdigo de la clase factora que se encarga de instanciar los pasos deductivos del esquema
	 * (StepFactory) a partir de una lista con los nombres de las clases de los diferentes pasos deductivos
	 * que se han de instanciar. Devuelve dicho cdigo.
	 * @param stepClassNames Nombre de las clases de pasos deductivos normales que tenemos.
	 * @param goalClassNames Nombre de las clases de pasos condicin de meta que tenemos.
	 * @return Cdigo de la clase factora que instancia los pasos deductivos.
	 */
	public String generateFactoryClassCode ( List stepClassNames , List goalClassNames )
	{
		StringBuffer sbInstanceSteps = new StringBuffer();
		
		sbInstanceSteps.append("/*Code Generated for File: " + "StepFactory.java*/\n");
		sbInstanceSteps.append("\n");
		
		/*Insertar la primera parte del cdigo, que es igual para cualquier esquema
		 * y tenemos almacenada en un fichero.*/
		try
		{
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"Factory.p1") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				sbInstanceSteps.append(linea);
				sbInstanceSteps.append("\n");
			}
		}
		catch ( IOException ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"Factory.p1:");
			ioe.printStackTrace();
		}
		
		/*
		 * Generar el mtodo instanceSteps, que vara para cada esquema.
		 */
		sbInstanceSteps.append ( "public static List instanceSteps ( Grammar g )\n{\n");
		sbInstanceSteps.append ( "List l = new ArrayList();\n");
		for ( int i = 0 ; i < stepClassNames.size() ; i++ )
			sbInstanceSteps.append ( stepClassNames.get(i) + ".getInstancesForGrammar(g,l);\n");
		sbInstanceSteps.append ( "return l;\n");
		sbInstanceSteps.append ( "}\n");
		
		/*
		 * Generar el mtodo instanceGoalConditions, que vara para cada esquema.
		 */
		sbInstanceSteps.append ( "public static List instanceGoalConditions ( Grammar g )\n{\n");
		sbInstanceSteps.append ( "List l = new ArrayList();\n");
		for ( int i = 0 ; i < goalClassNames.size() ; i++ )
			sbInstanceSteps.append ( goalClassNames.get(i) + ".getInstancesForGrammar(g,l);\n");
		sbInstanceSteps.append ( "return l;\n");
		sbInstanceSteps.append ( "}\n");
		
		/*
		 * Generar la ltima parte del cdigo, que es igual para todos los esquemas
		 * de anlisis sintctico, y est almacenada en un fichero.
		 */
		try
		{
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"Factory.p2") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				sbInstanceSteps.append(linea);
				sbInstanceSteps.append("\n");
			}
		}
		catch ( IOException ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"Factory.p2:");
			ioe.printStackTrace();
		}
		
		return ( sbInstanceSteps.toString() );
	}
	
	
	
	/**
	 * Genera el cdigo de la clase que se encarga de instanciar la agenda
	 * (Agenda) a partir de un fichero.
	 * This method is just like generateGrammarLoaderClassCode().
	 * @return Cdigo de la clase c
	 */
	public String generateAgendaLoaderClassCode (  )
	{
		StringBuffer sbLoader = new StringBuffer();
		
		sbLoader.append("/*Code Generated for File: " + "AgendaLoader.java*/\n"); //TODO: generalize into insertHeader() function
		sbLoader.append("\n");
		
		/*Insertar la primera parte del cdigo, que es igual para cualquier esquema
		 * y tenemos almacenada en un fichero.*/
		try
		{
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"AgendaLoader.p1") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				sbLoader.append(linea);
				sbLoader.append("\n");
			}
		}
		catch ( IOException ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"AgendaLoader.p1:"); //TODO: generalize into insertPart() function
			ioe.printStackTrace();
		}
		
		/*
		 * Generar el mtodo getAgenda().
		 */
		Class agendaClass = sn.getAgendaClass();
		String agendaClassName = agendaClass.getName();
		try
		{
			agendaClass.getConstructor(new Class[0]);
		}
		catch ( NoSuchMethodException nsme )
		{
			System.err.println("[WARNING] Agenda class " + agendaClassName + " has no default constructor as required, generated code will not compile.");
		}
		
		sbLoader.append("public static Agenda getAgenda ( )\n");
		sbLoader.append("{\n");
		sbLoader.append("  Agenda a = new " + agendaClassName + "();\n");
		sbLoader.append("  return a;\n");
		sbLoader.append("}\n");
		
		/*
		 * Generar la ltima parte del cdigo, que es igual siempre,
		 * y est almacenada en un fichero.
		 */
		try
		{
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"AgendaLoader.p2") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				sbLoader.append(linea);
				sbLoader.append("\n");
			}
		}
		catch ( IOException ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"AgendaLoader.p2:");
			ioe.printStackTrace();
		}
		
		return ( sbLoader.toString() );
	}
	
	
	/**
	 * Genera el cdigo de la clase que se encarga de instanciar el motor deductivo (DeductiveEngine).
	 * This method is just like generateAgendaLoaderClassCode().
	 * @return Cdigo de la clase c
	 */
	public String generateDeductiveEngineLoaderClassCode (  )
	{
		StringBuffer sbLoader = new StringBuffer();
		
		sbLoader.append("/*Code Generated for File: " + "DeductiveEngineLoader.java*/\n"); //TODO: generalize into insertHeader() function
		sbLoader.append("\n");
		
		/*Insertar la primera parte del cdigo, que es igual para cualquier esquema
		 * y tenemos almacenada en un fichero.*/
		try
		{
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"DeductiveEngineLoader.p1") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				sbLoader.append(linea);
				sbLoader.append("\n");
			}
		}
		catch ( IOException ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"DeductiveEngineLoader.p1:"); //TODO: generalize into insertPart() function
			ioe.printStackTrace();
		}
		
		/*
		 * Generar el mtodo getDeductiveEngine().
		 */
		String engineClassName = sn.getEngineClassName();
		
		sbLoader.append("public static DeductiveEngine getEngine ( Agenda a , List deductiveSteps , List goalConditions , ItemHandler itemHandler , StepHandler stepHandler )\n");
		sbLoader.append("{\n");
		sbLoader.append("  DeductiveEngine de = new " + engineClassName + "( a , deductiveSteps , goalConditions , itemHandler , stepHandler );\n");
		sbLoader.append("  return de;\n");
		sbLoader.append("}\n");
		
		/*
		 * Generar la ltima parte del cdigo, que es igual siempre,
		 * y est almacenada en un fichero.
		 */
		try
		{
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"DeductiveEngineLoader.p2") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				sbLoader.append(linea);
				sbLoader.append("\n");
			}
		}
		catch ( IOException ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"DeductiveEngineLoader.p2:");
			ioe.printStackTrace();
		}
		
		return ( sbLoader.toString() );
	}
	
	
	
	/**
	 * Genera el cdigo de la clase que se encarga de instanciar la gramtica
	 * (Grammar) a partir de un fichero.
	 * @return Cdigo de la clase c
	 */
	public String generateGrammarLoaderClassCode (  )
	{
		StringBuffer sbLoader = new StringBuffer();
		
		sbLoader.append("/*Code Generated for File: " + "GrammarLoader.java*/\n");
		sbLoader.append("\n");
		
		/*Insertar la primera parte del cdigo, que es igual para cualquier esquema
		 * y tenemos almacenada en un fichero.*/
		try
		{
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"GrammarLoader.p1") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				sbLoader.append(linea);
				sbLoader.append("\n");
			}
		}
		catch ( IOException ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"GrammarLoader.p1:");
			ioe.printStackTrace();
		}
		
		/*
		 * Generar el mtodo getGrammar().
		 */
		Class grammarClass = sn.getGrammarClass();
		String grammarClassName = grammarClass.getName();
		try
		{
			grammarClass.getConstructor(new Class[0]);
		}
		catch ( NoSuchMethodException nsme )
		{
			System.err.println("[WARNING] Grammar class " + grammarClassName + " has no default constructor as required, generated code will not compile.");
		}
		
		sbLoader.append("public static Grammar getGrammar ( File f ) throws IOException, IncorrectGrammarFormatException\n");
		sbLoader.append("{\n");
		sbLoader.append("  Grammar g = new " + grammarClassName + "();\n");
		sbLoader.append("  g.initFromFile(f);\n");
		sbLoader.append("  return g;\n");
		sbLoader.append("}\n");
		
		/*
		 * Generar la ltima parte del cdigo, que es igual siempre,
		 * y est almacenada en un fichero.
		 */
		try
		{
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"GrammarLoader.p2") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				sbLoader.append(linea);
				sbLoader.append("\n");
			}
		}
		catch ( IOException ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"GrammarLoader.p2:");
			ioe.printStackTrace();
		}
		
		return ( sbLoader.toString() );
	}
	
	
	
	/**
	 * Genera el cdigo correspondiente a todos los pasos de un esquema de anlisis
	 * sintctico dado por el SchemaNode que recibimos como parmetro.
	 * @param sn Nodo raz del rbol que representa a un esquema de anlisis
	 * sintctico.
	 * @return 
	 */
	public String generateCode ()
	{	
		StringBuffer sb = new StringBuffer("");
		List itemSearchSkeletonTrees = new ArrayList();
		List stepSearchSkeletonTrees = new ArrayList();
		List existenceSkeletonTrees = new ArrayList();
		List stepClassNames = new ArrayList(); //nombres de las clases de los pasos normales, para instanciarlos en la factory
		List goalClassNames = new ArrayList(); //nombres de las clases de los pasos meta, para instanciarlos en la factory
		
		//Generar el cdigo para los pasos deductivos normales, y obtener:
		//	* Descriptores de bsqueda para la generacin del ItemHandler,
		//	* Nombres de las clases para la generacin del cdigo que instancia los pasos.
		for ( int nstep = 0 ; nstep < sn.numChildren() ; nstep++ )
		{
			DeductiveStepNode dsn = sn.getChild(nstep);
			if ( !dsn.isGoalStep() )
			{
				DeductiveStepCodeGenerator dscg = new DeductiveStepCodeGenerator(dsn);
				
				if ( VERBOSE_OUTPUT )
					System.out.println("Generating code for deductive step " + (nstep+1) + "...");
				String thisStepCode = dscg.generateCode ( itemSearchSkeletonTrees , stepSearchSkeletonTrees , existenceSkeletonTrees );
				String stepCodeFileName = outputDirectory+"src"+File.separator+"schema"+File.separator+"step"+File.separator+ dscg.getClassName()+".java";
	
				sb.append ( thisStepCode );	
				try
				{
					dumpToFile ( stepCodeFileName , thisStepCode );
					stepClassNames.add( dscg.getClassName() );
					if ( VERBOSE_OUTPUT )
						System.out.println("...generated [file: " + stepCodeFileName + ", size: " + thisStepCode.length() + " characters]");
				}
				catch ( IOException ioe )
				{
					ioe.printStackTrace();
				}
			}
		}
			
		//Generar el cdigo para la clase ItemHandler que maneja los items.
		if ( VERBOSE_OUTPUT )
			System.out.println("Generating code for item indexing...");
		String poolCode = generateItemHandlerCode ( itemSearchSkeletonTrees , existenceSkeletonTrees );
		sb.append ( poolCode );	
		String ItemHandlerCodeFileName = outputDirectory+"src"+File.separator+"schema"+File.separator+"item"+File.separator+"ItemHandler.java";
		try
		{
			dumpToFile ( ItemHandlerCodeFileName , poolCode );
			if ( VERBOSE_OUTPUT )
				System.out.println("...generated [file: " + ItemHandlerCodeFileName + ", size: " + poolCode.length() + " characters]");
		}
		catch ( IOException ioe )
		{
			ioe.printStackTrace();
		}
		
		//Generar el cdigo para la clase StepHandler que indexa los pasos.
		if ( VERBOSE_OUTPUT )
			System.out.println("Generating code for deductive step indexing...");
		String StepHandlerCode = generateStepHandlerCode ( stepSearchSkeletonTrees );
		sb.append ( StepHandlerCode );
		String StepHandlerCodeFileName = outputDirectory+"src"+File.separator+"schema"+File.separator+"step"+File.separator+"StepHandler.java";
		try
		{
			dumpToFile ( StepHandlerCodeFileName , StepHandlerCode );
			if ( VERBOSE_OUTPUT )
				System.out.println("...generated [file: " + StepHandlerCodeFileName + ", size: " + StepHandlerCode.length() + " characters]");
		}
		catch ( IOException ioe )
		{
			ioe.printStackTrace();
		}
		
		//Generar el cdigo para los pasos deductivos de meta, que no intervienen
		//en el algoritmo pero determinan si un item dado es meta o no.
		for ( int nstep = 0 ; nstep < sn.numChildren() ; nstep++ )
		{
			DeductiveStepNode dsn = sn.getChild(nstep);
			if ( dsn.isGoalStep() )
			{
				DeductiveStepCodeGenerator dscg = new DeductiveStepCodeGenerator(dsn);
				
				if ( VERBOSE_OUTPUT )
					System.out.println("Generating code for goal condition " + (nstep+1) + "...");
				String thisStepCode = dscg.generateCode ( new ArrayList() , new ArrayList() , new ArrayList() );
				String stepCodeFileName = outputDirectory+"src"+File.separator+"schema"+File.separator+"step"+File.separator+ dscg.getClassName()+".java";
				sb.append ( thisStepCode );	
				try
				{
					dumpToFile ( stepCodeFileName , thisStepCode );
					goalClassNames.add( dscg.getClassName() );
					if ( VERBOSE_OUTPUT )
						System.out.println("...generated [file: " + stepCodeFileName + ", size: " + thisStepCode.length() + " characters]");
				}
				catch ( IOException ioe )
				{
					ioe.printStackTrace();
				}
			}
		}
		
		//Generar el cdigo para instanceSteps
		if ( VERBOSE_OUTPUT )
			System.out.println("Generating code for step factory...");
		String factoryCode = generateFactoryClassCode ( stepClassNames , goalClassNames );
		sb.append ( factoryCode );
		String factoryCodeFileName = outputDirectory+"src"+File.separator+"schema"+File.separator+"step"+File.separator+"StepFactory.java";
		try
		{
			dumpToFile ( factoryCodeFileName , factoryCode );
			if ( VERBOSE_OUTPUT )
				System.out.println("...generated [file: " + factoryCodeFileName + ", size: " + factoryCode.length() + " characters]");
		}
		catch ( IOException ioe )
		{
			ioe.printStackTrace();
		}
		
		//Generar el cdigo para grammar loader
		if ( VERBOSE_OUTPUT )
			System.out.println("Generating code for grammar loader...");
		String loaderCode = generateGrammarLoaderClassCode();
		sb.append ( loaderCode );
		String loaderCodeFileName = outputDirectory+"src"+File.separator+"grammar"+File.separator+"GrammarLoader.java";
		try
		{
			dumpToFile ( loaderCodeFileName , loaderCode );
			if ( VERBOSE_OUTPUT )
				System.out.println("...generated [file: " + loaderCodeFileName + ", size: " + loaderCode.length() + " characters]");
		}
		catch ( IOException ioe )
		{
			ioe.printStackTrace();
		}
		
		//Generar el cdigo para engine loader
		if ( VERBOSE_OUTPUT )
			System.out.println("Generating code for deductive engine loader...");
		loaderCode = generateDeductiveEngineLoaderClassCode();
		sb.append ( loaderCode );
		loaderCodeFileName = outputDirectory+"src"+File.separator+"engine"+File.separator+"DeductiveEngineLoader.java";
		try
		{
			dumpToFile ( loaderCodeFileName , loaderCode );
			if ( VERBOSE_OUTPUT )
				System.out.println("...generated [file: " + loaderCodeFileName + ", size: " + loaderCode.length() + " characters]");
		}
		catch ( IOException ioe )
		{
			ioe.printStackTrace();
		}
		
		//Generar el cdigo para agenda loader
		if ( VERBOSE_OUTPUT )
			System.out.println("Generating code for agenda loader...");
		loaderCode = generateAgendaLoaderClassCode();
		sb.append ( loaderCode );
		loaderCodeFileName = outputDirectory+"src"+File.separator+"agenda"+File.separator+"AgendaLoader.java";
		try
		{
			dumpToFile ( loaderCodeFileName , loaderCode );
			if ( VERBOSE_OUTPUT )
				System.out.println("...generated [file: " + loaderCodeFileName + ", size: " + loaderCode.length() + " characters]");
		}
		catch ( IOException ioe )
		{
			ioe.printStackTrace();
		}
		
		//Copiar el cdigo que es comn a todos los "parsers" generados: manejo general
		//de items, elementos de items, etc.
		if ( VERBOSE_OUTPUT )
			System.out.println("Copying common files...");
		try
		{
			File[] thingsToCopy = (new File("common")).listFiles();
			for ( int i = 0 ; i < thingsToCopy.length ; i++ )
			{
				if ( !thingsToCopy[i].getName().equalsIgnoreCase("parts") )
					FileCopy.copyFile(thingsToCopy[i],new File(outputDirectory,thingsToCopy[i].getName()));
			}
			if ( VERBOSE_OUTPUT )
				System.out.println("...copy successfully completed.");
		}
		catch ( IOException ioe )
		{
			ioe.printStackTrace();
		}
		
		return new String(sb);
	}
	
	/**
	 * Escribe una cadena dada en el fichero de nombre dado.
	 * @param filename Fichero en el que se desea escribir. Puede estar dado en forma
	 * de path relativo, en cuyo caso si no existe el directorio indicado se crear.
	 * @param s Contenido que se escribir al fichero.
	 * @throws FileNotFoundException Si el fichero no se puede abrir.
	 * @throws IOException Si hay algn problema de entrada/salida.
	 */
	private void dumpToFile ( String filename , String s ) throws FileNotFoundException , IOException
	{
		File theFile = new File(filename);
		if ( theFile.getParentFile() != null )
			theFile.getParentFile().mkdirs(); //construye los directorios necesarios
		OutputStreamWriter theWriter = new OutputStreamWriter(new FileOutputStream(theFile));
		theWriter.write(s);
		theWriter.close();
	}
	
	/**
	 * Genera el cdigo correspondiente a la clase ItemHandler, que es la que
	 * maneja los items y proporciona acceso rpido a ellos.
	 * @param ofSkeletonTrees Lista que contiene los "skeleton trees" que describen
	 * las diferentes maneras en que se podrn hacer peticiones de items a la
	 * ItemHandler. El cdigo generado deber dar servicio a estos "skeleton trees"
	 * de alguna manera.
	 * @param existenceSkeletonTrees Lista que contiene los "skeleton trees" para
	 * las peticiones de tipo "exists" (para saber de la existencia o no de items
	 * en la ItemHandler).
	 * @return Cadena con el cdigo generado.
	 * @see ApplySpecificationVisitor, SearchDescriptorNode
	 */
	public String generateItemHandlerCode ( List itemSearchSkeletonTrees , List existenceSkeletonTrees )
	{
		SearchDescriptorNode[] searchSkelTrees = (SearchDescriptorNode[]) itemSearchSkeletonTrees.toArray(new SearchDescriptorNode[0]);	
		SearchDescriptorNode[] existenceSkelTrees = (SearchDescriptorNode[]) existenceSkeletonTrees.toArray(new SearchDescriptorNode[0]);
		StringBuffer poolSb = new StringBuffer("");
		ItemHandlerCodeGenerator ipcg = new ItemHandlerCodeGenerator();
		poolSb.append("/*Code Generated for File: ItemHandler.java*/\n");
		poolSb.append ( ipcg.generateCode(searchSkelTrees,existenceSkelTrees) );
		return new String(poolSb);
	}
	
	/**
	 * Genera el cdigo correspondiente a la clase StepHandler, que es la que
	 * proporciona acceso rpido a los pasos deductivos.
	 * @param ofSkeletonTrees Lista que contiene los "skeleton trees" que describen
	 * las diferentes maneras en que se podrn hacer peticione sde pasos deductivos
	 * a la StepHandler. El cdigo generado deber dar servicio a estos "skeleton trees"
	 * de alguna manera.
	 * @return Cadena con el cdigo generado.
	 * @see ApplySpecificationVisitor, SearchDescriptorNode
	 */
	public String generateStepHandlerCode ( List ofSkeletonTrees )
	{
		SearchDescriptorNode[] skelTrees = (SearchDescriptorNode[]) ofSkeletonTrees.toArray(new SearchDescriptorNode[0]);	
		StringBuffer poolSb = new StringBuffer("");
		StepHandlerCodeGenerator spcg = new StepHandlerCodeGenerator();
		poolSb.append("/*Code Generated for File: StepHandler.java*/\n");
		poolSb.append ( spcg.generateCode(skelTrees) );
		return new String(poolSb);
	}
		
}
