/*
 * Created at regulus on 06-nov-2004 17:29:14
 * as file IndexDescriptor.java on package generator
 */
package generator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * 
 * <p>Cada instancia de esta clase describe un ndice sobre algn tipo de especificacin de bsqueda
 * que pueden pasar a la clase ItemHandler, que debemos generar, para que sta devuelva
 * items.</p>
 * <p>Cada uno de estos ndices puede incluir varios nodos del rbol e indexar de
 * formas distintas sobre cada uno. Por ejemplo, puede hacer un ndice array sobre
 * un nodo y hash sobre otro (y lo que tendramos en la clase ItemHandler sera un
 * array de tablas hash para acceso rpido a los items).</p>
 * <p>Para representar esto, se lleva una lista de ndices atmicos (clase AtomicIndex),
 * cada uno de ellos contiene la direccin en el rbol de uno de los nodos por los que se 
 * indexa y el tipo de indexacin que se lleva a cabo con ese nodo.</p>
 * <p>Las direcciones que se utilizan para este propsito son listas de Integer, de manera
 * que el nodo raz es la lista vaca, sus hijos son 0, 1, 2, 3, ...; los hijos del nodo
 * k son (k,0), (k,1), ... , etc.</p>
 * <p>Esta clase implementa el interfaz Comparable porque nos interesar tener los
 * ndices ordenados, primero los ms restrictivos, cuando creemos la lista de ndices.</p>
 * <p>Created 12-ago-2004 17:31:09</p>
 * @author carlos
 */
public class IndexDescriptor implements Comparable
{
    
    /**
     * Nombre del ndice.
     */
    private String name;
    
    /**
     * rboles que describen las bsquedas que utilizarn este mtodo de
     * indexacin.
     */
    private List associatedSkeletons = new ArrayList(); //of SearchDescriptorNode
    
    /**
     * Descripcin de de qu manera se indexa cada uno de los atributos del
     * skeleton tree.
     */
    private List atomicIndices = new ArrayList(); //of AtomicIndex
    
    /**
     * Compara este ndice con otro en funcin de la cantidad de ndices atmicos
     * que posea (i.e. la cantidad de "cosas" por las que indexe)
     * @param o Otro ndice con el que comparamos.
     * @return 1, 0 o -1.
     */
    public int compareTo ( Object o )
    {
    	IndexDescriptor id = (IndexDescriptor) o;
    	if ( this.atomicIndices.size() > id.atomicIndices.size() ) return -1;
    	else if ( this.atomicIndices.size() < id.atomicIndices.size() ) return +1;
    	else return 0;
    }
    
    /**
     * Nos dice si un descriptor de ndice es igual a otro. Se considera que dos
     * descriptores de ndice son iguales si tienen los mismos ndices atmicos,
     * en cualquier orden. (no se consideran influyentes el nombre y los esqueletos
     * asociados).
     */
    public boolean equals ( Object o )
    {
    	if ( !(o instanceof IndexDescriptor) ) return false;
    	IndexDescriptor id = (IndexDescriptor) o;
    	if ( this.atomicIndices.size() != id.atomicIndices.size() ) return false;
    	//comparacin sencilla. Mirar si todos los ndices de this estn en id.
    	for ( int i = 0 ; i < this.atomicIndices.size() ; i++ )
    	{
    		boolean found = false;
    		for ( int j = 0 ; j < id.atomicIndices.size() ; j++ )
    		{
    			if ( this.atomicIndices.get(i).equals(id.atomicIndices.get(j)) )
    			{
    				found = true;
    				break;
    			}
    		}
    		if ( !found ) return false;
    	}
    	System.out.println(this + " EQUALS " + id);
    	return true;
    }
    
    
    
    /**
     * Construye un ndice vaco (es decir, no se indexa por nada, sera simplemente
     * una lista de items para recorrer de manera secuencial) de nombre dado.
     * @param name Nombre que tomar el ndice.
     */
    public IndexDescriptor ( String name )
    {
        this.name = name;
    }

    /**
     * Describe un ndice para los rboles esqueleto dados, con el nombre dado, y que
     * est formado por los ndices atmicos (vase javadoc de la clase) dados.
     * @param name Nombre del ndice.
     * @param associatedSkeletons Descriptores de las bsquedas a los que se aplicar este ndice.
     * @param atomicIndices ndices atmicos que representan cada uno de los nodos
     * del rbol de item que se indexan y de qu manera. Ntese que el mismo objeto
     * AtomicIndex no debe aparecer en dos IndexDescriptor distintos, dado que el
     * AtomicIndex establece una referencia al IndexDescriptor padre.
     */
    public IndexDescriptor ( String name , List associatedSkeletons , List atomicIndices )
    {
        this.name = name;
        this.associatedSkeletons = associatedSkeletons;
        this.atomicIndices = atomicIndices;
        for ( int i = 0 ; i < atomicIndices.size() ; i++ )
        {
            ((AtomicIndex)atomicIndices.get(i)).setParent(this);
        }
    }
    
    /**
     * Agrega un ndice atmico al descriptor de ndice.
     * Ntese que no se debe agregar el mismo objeto AtomicIndex a dos IndexDescriptor
     * diferentes, dado que el AtomicIndex establece una referencia al padre.
     * @param ai ndice atmico (i.e. ndice para un nodo concreto) a agregar.
     */
    public void addAtomicIndex( AtomicIndex ai )
    {
        atomicIndices.add(ai);
        ai.setParent(this);
    }
    
    /**
     * Asocia este descriptor de ndice a un determinado rbol de esqueletos,
     * que describe una estructura de esqueleto que podran pasar al ItemHandler y que
     * se resolvera con este ndice.
     * Slo se aade el rbol de esqueleto si no estaba ya (este add no crea duplicados)
     * @param stn rbol de esqueleto a aadir.
     */
    public void addSearchDescriptor ( SearchDescriptorNode stn )
    {
    	if ( !associatedSkeletons.contains(stn) )
    		associatedSkeletons.add(stn);
    	System.out.println("CURRENT: " + associatedSkeletons.get(0));
    	System.out.println("INCOMING: " + stn);
    	System.out.println("NODES: " + associatedSkeletons.size());
    }
    
    /**
     * Devuelve la lista de ndices atmicos.
     * @return ndices atmicos asociados a este descriptor.
     */
    public List getAtomicIndexes()
    {
        return atomicIndices;
    }
    
    /**
     * Devuelve la lista de descriptores de bsqueda asociados.
     * @return Una lista de SearchDescriptorNodes con los rboles que describen las bsquedas
     * a los que puede dar servicio esta estructura de ndices.
     */
    public List getAssociatedSearchDescriptors()
    {
        return associatedSkeletons;
    }
    
    /**
     * Devuelve el nombre asociado a este ndice.
     * @return El nombre del ndice, que usaremos como nombre de variable.
     */
    public String getName()
    {
        return name;
    }
    
    /**
     * Obtiene una declaracin parcial para el ndice. (tratamiento de nulos)
     */
    private String getPartialAttributeDeclaration ( int initialIndex )
    {
        
        List indices = this.getAtomicIndexes();
        
        StringBuffer decltype = new StringBuffer("");
        
        //este flag indica que an no se ha encontrado un tipo que nos impida continuar
        //aadiendo especificadores de tipo.
        //i.e. los arrays se pueden concatenar ([][]...) pero las HashMaps no, no hay
        //un tipo "tabla hash de tabla hash" al no haber genricos.
        boolean noBaseType = true;
        
        for ( int i = initialIndex ; i < indices.size() ; i++ )
        {
            AtomicIndex ai = (AtomicIndex) indices.get(i);
            
            if ( ai.getIndexType() == AtomicIndex.ARRAY_INDEX )
            {
                if ( noBaseType )
                    decltype.append("[]");
                else
                    decltype.append("/*of array*/ ");
            }
            else if ( ai.getIndexType() == AtomicIndex.HASH_INDEX )
            {
                if ( noBaseType )
                {
                    noBaseType = false;
                	decltype.insert(0,"HashMap");
                }
                else
                    decltype.append("/*of HashMap*/");
            }
        }
        
        if ( noBaseType )
            decltype.insert(0,"List /*of Object[]*/ ");
        else
            decltype.append("/*of List of Object[]*/ ");
        
        //decltype tiene el valor con que hemos de declarar el resultado parcial.
        //slo contiene el nombre de la clase (no es una declaracin completa con
        //nombre de variable).
        
        return decltype.toString();
        
    }
    
    /**
     * Obtiene una inicializacin parcial para el ndice. (tratamiento de nulos)
     */
    private String getPartialAttributeInitialization ( int initialIndex )
    {
        
        List indices = this.getAtomicIndexes();
        
        //llamada a constructor que va despus del new.
        StringBuffer constructor = new StringBuffer("");
        
        //este flag indica que an no se ha encontrado un tipo que nos impida continuar
        //aadiendo especificadores de tipo.
        //i.e. los arrays se pueden concatenar ([][]...) pero las HashMaps no, no hay
        //un tipo "tabla hash de tabla hash" al no haber genricos.
        boolean noBaseType = true;
        
        for ( int i = initialIndex ; i < indices.size() ; i++ )
        {
            AtomicIndex ai = (AtomicIndex) indices.get(i);
            
            if ( ai.getIndexType() == AtomicIndex.ARRAY_INDEX )
            {
                if ( noBaseType )
                {
                    String lenSpec = ai.getIndexedNodeClass().getName() + ".getArrayLength()";
                    constructor.append("["+lenSpec+"]");
                }
            }
            else if ( ai.getIndexType() == AtomicIndex.HASH_INDEX )
            {
                if ( noBaseType )
                {
                    noBaseType = false;
                	constructor.insert(0,"HashMap");
                }
            }
        }
        
        if ( noBaseType )
            constructor.insert(0,"ArrayList /*of Object[]*/ ");

        if ( !constructor.toString().endsWith("]") ) constructor.append("()");
        
        return /*id.getName() +*/ "new " + constructor.toString() ;
        
    }
    
    
    private List partialResults = new ArrayList(); /*of PartialAccessResult*/
    
    /**
     * @see getIndexUsageCode()
     * @return La lista de resultados parciales, til por si hay estructuras nulas
     * en un camino a travs del ndice, tal y como se describe en la documentacin
     * de getIndexUsageCode().
     */
    public List getPartialIndexUsageCodeResults()
    {
    	return partialResults;
    }
    
    class PartialAccessResult
	{
    	
    	private String accessCode;
    	private String initCode;
    	private String declarationCode;
    	
    	public PartialAccessResult ( String accessCode , String initCode , String declarationCode )
    	{
    		this.accessCode = accessCode;
    		this.initCode = initCode;
    		this.declarationCode = declarationCode;
    	}
    	
    	public String getAccessCode(){return accessCode;}
    	public String getInitializerCode(){return initCode;}
    	public String getDeclarationCode(){return declarationCode;}
    	public void replaceAccessCode ( String regex1 , String regex2 )
    	{
    		accessCode = accessCode.replaceFirst(regex1,regex2);
    	}
    	
	}
    
    /**
     * <p>Obtiene el cdigo para consultar un ndice por medio de un tem o especificacin de bsqueda
     * (en el cdigo generado) de nombre objectName. Por ejemplo, si es un ndice de
     * array por la tercera componente del item, devolver algo como
     * nuestroIndice[nuestroItem[2]];</p> 
     * <p>Esto lo hace de forma lineal, por ejemplo para un ndice (array, hash, array) podra
     * dar algo como</p>
     *           <em>((List[])((HashMap)((HashMap[])index)[i]).get(j))[k]</em>
     * <p>(el cdigo de esta funcin siempre debera devolver algo de tipo List).</p>
     * <p>Las claves de indexacin i, j y k se obtienen de una tabla de conversin de direcciones,
     * que a partir de una direccin en el rbol DESCRIPTOR de bsqueda nos devuelve la cadena
     * que identifica al objeto en la ESPECIFICACIN de bsqueda del cdigo (ntese la diferencia. El descriptor
     * de bsqueda puede tener Sequences, etc. y el nodo nmero 2, por ejemplo, podra
     * ser el nodo i+1 en el cdigo).</p>
     * @param addressConverter Una tabla hash que indique cmo las direcciones en el descriptor
     * de bsqueda que indican los nodos por los que se indexa deben traducirse a cdigo.
     * Por ejemplo, una entrada sencilla podra ser {0,1} -> ((Object[])((Object[])item)[0])[1];
     * pero las entradas se complicarn si aparecen secuencias, con la presencia de variables.
     * @return El cdigo para usar el ndice, tal y como se ha dicho. Ntese, sin embargo,
     * que este cdigo podra tirar una NullPointerException si alguna de las estructuras
     * intermedias nos devuelve un objeto nulo. Para poder arreglar fcilmente esto sin
     * necesidad de capturar una NullPointerException (cosa poco eficiente pues se crea
     * un objeto), como efecto lateral, este mtodo crea una lista partialResults con
     * cadenas de tipo algo = new otroAlgo(), donde la parte izquierda de la igualdad
     * es cada una de las etapas intermedias del uso del ndice, y la parte izquierda
     * junto con la derecha nos dan la asignacin que habra que hacer, en caso de que
     * la izquierda fuese nula, para convertirla en una estructura vaca (cosa til
     * si se quieren aadir elementos). Esta lista se obtiene, para la ltima
     * llamada a este mtodo hecha, con getPartialIndexUsageCodeResults.
     * @param initialization [04-10-27] Si est a true, en lugar del cdigo para obtener
     * lo que haya en el ndice, devuelve el cdigo para inicializar la posicin
     * destino del ndice a la lista vaca.
     */
    public String getIndexUsageCode ( HashMap addressConverter , boolean initialization )
    {
       
        
        partialResults = new ArrayList(); //of StringBuffer
        boolean addPartialResult = false;
        String tempPartialInitterCode="ERR";
        
        
        
        //empezamos con el nombre del ndice
        StringBuffer result = new StringBuffer(this.getName());
        
        //buffer temporal
        StringBuffer accumulatedArrayReferences = new StringBuffer("");
    
        //recorremos los ndices atmicos
        List atomicIndices = this.getAtomicIndexes();
        boolean typeCastNecessary = false; //a partir de usar una Collection harn falta casts.
        boolean typeCastOpen = false; //si se ha abierto un cast que no se ha cerrado
        int arraydims = 1; //dimensiones de array para el typecast
        int previousdims = 1; //valor anterior de arraydims
        int currentIndexType = AtomicIndex.ARRAY_INDEX;; //tipo de ndice procesado
        for ( int i = 0 ; i < atomicIndices.size() ; i++ )
        {
            addPartialResult = false;
            AtomicIndex ai = (AtomicIndex) atomicIndices.get(i);
            switch ( currentIndexType = ai.getIndexType() )
            {
                case AtomicIndex.ARRAY_INDEX:
                {
                    //insertar type cast si es necesario
                    if ( typeCastNecessary )
                    {
                        //mirar de qu tipo va a ser el array.
                        
                        //si ya no hay ms ndices, ser de Lists.
                        if ( i == atomicIndices.size()-1 )
                        {
                            StringBuffer temp = new StringBuffer("");
                            temp.append("((List");
                            for ( int r = 0 ; r < arraydims ; r++  )
                                temp.append("[]");
                            temp.append(")");
                            result.insert(0,temp);
                            result.append(")");
                            
                            //if ( arraydims == 1 )
                                addPartialResult = true;
                            
                            previousdims = arraydims;
                            arraydims = 1;
                            
                            
                        }
                        else
                        {
                            AtomicIndex next = (AtomicIndex) atomicIndices.get(i+1);
                            switch ( next.getIndexType() )
                            {
                                //si despus viene un array, ser un array de arrays;
                                //pero un array de arrays lo estamos representando como
                                //array multidimensional, as que incrementamos un contador
                                //arraydims y acumulamos hasta la siguiente iteracin.
                                case AtomicIndex.ARRAY_INDEX:
                                {
	                                //el siguiente nodo ser el que establezca el typecast.
	                                //Vamos a tener algo de tipo (tal)a[i][j]. => es al
	                                //procesar la [j] que se pone el typecast, no al procesar
	                                //la i.
                                	
                                	//if ( arraydims == 1 )
	                                //	addPartialResult = true;
                                	
                                	previousdims = arraydims;
	                                arraydims++; //esto le indicar que ha de poner [][].	
	                                	                                
                             
	                                break; //inner case switch                                
                                }
                                //si despus viene una HashMap, el array ser de HashMaps.
                                case AtomicIndex.HASH_INDEX:
                                {
                                    //como cuando no haba ndice pero con HashMap
                                    //en vez de list.
                                    StringBuffer temp = new StringBuffer("");
                                    temp.append("((HashMap");
                                    for ( int r = 0 ; r < arraydims ; r++  )
                                        temp.append("[]");
                                    temp.append(")");
                                    result.insert(0,temp);
                                    result.append(")");
                                    

                                        addPartialResult = true;
                                    
                                        previousdims = arraydims;
                                    arraydims = 1;
                                    break; //case switch
                                } //case
                            } //next index case switch
                        } //elif there is a next index
                    } //if type cast is necessary
                    


                                        
                    //case array index
                    
                    //ahora addPartialResult va asociado siempre a arraydims == 1
                    
                    if ( i>0 && addPartialResult )
                    {
                    	//partial result initter string
                    	AtomicIndex anterior = (AtomicIndex) atomicIndices.get(i-1);
                    	String resultInitter="";
                    	resultInitter = tempPartialInitterCode.replaceAll("\\Q%TEMP%\\E",this.getPartialAttributeInitialization(i-(previousdims-1))+"/*INI"+(i-(previousdims-1))+"type"+((AtomicIndex)atomicIndices.get(i-(previousdims-1))).getIndexType()+"*/");
                    	partialResults.add ( new PartialAccessResult(result.toString(),resultInitter,this.getPartialAttributeDeclaration(i-(previousdims-1))) );
                    }                    	
                    
                    if ( arraydims == 1 )
                    {
                    	//indexaciones de array anteriores
                    	if ( accumulatedArrayReferences.length() > 0 )
                    	{
                    		result.append(accumulatedArrayReferences);
                    		accumulatedArrayReferences = new StringBuffer("");
                    	}
                    	
                    	//indexacin de array actual
                    	result.append("[");
                    	result.append(addressConverter.get(ai.getTreeNodeAddress()));
                    	result.append(".getArrayIndex()");
                    	result.append("]");
                    	
                    	//cdigo de inicializacin
                        tempPartialInitterCode = result.toString();
                        tempPartialInitterCode = tempPartialInitterCode + " = %TEMP%";
                    	
                    }
                    
                    else if ( arraydims > 1 ) 
                    {
                    	/*
                    	 *an no insertamos en el resultado la indexacin del array ([]), pues
                    	 *lo haremos al acabar de recorrer todos los ndices tipo array para que quede
                    	 *bien el "type cast".
                    	*/
                    	accumulatedArrayReferences.append("[");
                    	accumulatedArrayReferences.append(addressConverter.get(ai.getTreeNodeAddress()));
                    	accumulatedArrayReferences.append(".getArrayIndex()");
                    	accumulatedArrayReferences.append("]");
                    }
                  
                    
                    
                    break;
                    
                } //case (array index)
                case AtomicIndex.HASH_INDEX:
                {
                    //insertar type cast si es necesario
                    if ( typeCastNecessary )
                    {
                        result.insert(0,"((HashMap)");
                        result.append(")");
                    }      
                    else
                    {
                        result.insert(0,"(");
                        result.append(")");
                    }
                    
                    if ( i > 0 )
                    {
                    	//partial result initter string
                    	AtomicIndex anterior = (AtomicIndex) atomicIndices.get(i-1);
                    	String resultInitter="";
                    	resultInitter = tempPartialInitterCode.replaceAll("\\Q%TEMP%\\E",this.getPartialAttributeInitialization(i));
                    	partialResults.add ( new PartialAccessResult(result.toString(),resultInitter,this.getPartialAttributeDeclaration(i)) );
                    }
                    	
                    tempPartialInitterCode = result.toString();
                    
                    result.append(".get(");
                    result.append(addressConverter.get(ai.getTreeNodeAddress()));
                    result.append(".getHashKey()");
                    result.append(")");
                    
                    tempPartialInitterCode = tempPartialInitterCode + ".put(";
                    tempPartialInitterCode += addressConverter.get(ai.getTreeNodeAddress());
                    tempPartialInitterCode += ".getHashKey()";
                    tempPartialInitterCode += ", %TEMP% )"; 
                    
                    typeCastNecessary = true; //pues esto nos devuelve un Object
                    
                    addPartialResult = true;
                    
                    //si ya no hay ms ndices, tendremos que hacer cast a List.
                    if ( i == atomicIndices.size()-1 )
                    {
                        result.insert(0,"(List)");
                    }                 
                    
                    break;
                } //case (hash index)
            } //case switch
            
            //if ( addPartialResult )
                //partialResults.add( new StringBuffer(result.toString()) + this.getPartialAttributeInitialization(i+1) );
            	//partialResults.add ( new PartialAccessResult(result.toString(),this.getPartialAttributeInitialization(i+1),this.getPartialAttributeDeclaration(i+1)) );
            	
        } //for loop
        
        
        if ( initialization )
        {
        	//si el ndice es array, simplemente igualar a new ArrayList().
        	if ( currentIndexType == AtomicIndex.ARRAY_INDEX )
        	{
        		result.append ( " = new ArrayList();\n");
        	}
        	//si es hash, usar el temp initter.
        	else if ( currentIndexType == AtomicIndex.HASH_INDEX )
        	{
            	//AtomicIndex anterior = (AtomicIndex) atomicIndices.get(i-1);
            	//String resultInitter="";
            	String resultInitter = tempPartialInitterCode.replaceAll("\\Q%TEMP%\\E","new ArrayList()");
            	//partialResults.add ( new PartialAccessResult(result.toString(),resultInitter,this.getPartialAttributeDeclaration(i)) );
            	resultInitter+=(";\n");
            	result = new StringBuffer(resultInitter);
        	}	
        }
        
        
        return result.toString();
        
    }
    
    
    /**
     * Proporciona una representacin legible del ndice en forma de cadena.
     */
    public String toString()
    {
    	StringBuffer sb = new StringBuffer("Index on ");
    	for ( Iterator it = atomicIndices.iterator() ; it.hasNext() ; )
    	{
    		AtomicIndex ai = (AtomicIndex) it.next();
    		sb.append( ( ai.toString() ) );
    	}
    	sb.append(" associated to search descriptors ");
    	for ( Iterator it = associatedSkeletons.iterator() ; it.hasNext() ; )
    	{
    		SearchDescriptorNode stn = (SearchDescriptorNode) it.next();
    		sb.append(stn.toString());
    	}
    	sb.append(".");
    	return sb.toString();
    }
}

