/*
 * Created on 12-ago-2004 16:54:42
 *
 */
package generator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.*;

/**
 * <p>Genera el cdigo para la clase que se hace cargo de los items y que se encarga
 * de proporcionar acceso rpido a ellos mediante ndices cuando sea posible.</p>
 * <p>Created 12-ago-2004 16:54:42</p>
 * @author carlos
 */
public class ItemHandlerCodeGenerator
{

	/**Indica si se activar la indexacin de existencia.*/
	private boolean EXISTENCE_INDEXING = true;
	
    /**
     * <p>Este mtodo, sabiendo el tipo de bsquedas que se van a utilizar para buscar
     * items en el algoritmo de parsing, toma la decisin de qu ndices crear.</p>
     * Cada uno de los ndices viene descrito por una estructura IndexDescriptor.
     * @param namePreffix Prefijo que se dar a los nombres de los ndices.
     * @param defaultIndexName Nombre del ndice por defecto (ndice vaco).
     * @param skeletonStructures Descriptores de bsqueda que describen los tipos de peticiones
     * que pueden pasar a la ItemHandler para recuperar items.
     * @return Descriptores de los ndices elegidos.
     */
    public List getSuitableIndices ( String namePreffix , String defaultIndexName , SearchDescriptorNode[] skeletonStructures )
    {
        
        ArrayList descriptors = new ArrayList();
        
        //En primer lugar, siempre habr un ndice "elemental", el ndice vaco, es decir,
        //simplemente la lista de items. No est asociado a ningn descriptor concreto;
        //pero ser al que se recurra cuando no haya otra opcin.
        
        IndexDescriptor defaultIndex = new IndexDescriptor (defaultIndexName);
        
        for ( int i = 0 ; i < skeletonStructures.length ; i++ )
        {
            
            SearchDescriptorNode skelTree = skeletonStructures[i];
            
            IndexDescriptor id = new IndexDescriptor(namePreffix+i);
            
            //creamos todos los descriptores de ndices posibles, y luego eliminaremos
            //los innecesarios, costosos, redundantes, etc. (de momento no, los dejamos
            //todos)
            
            addAllPossibleIndexes ( id , new ArrayList() , skelTree );
            
            /*
             * Si se quisiera aadir cdigo para quitar los ndices atmicos sobrantes o redundantes,
             * debera hacerse aqu.
             */
            
            //De momento, simplemente, quitamos los iguales (repetidos).
            
            //equality check
            boolean exists = false;
            for ( int j = 0 ; j < descriptors.size() ; j++ )
            {
            	IndexDescriptor existingDescriptor = (IndexDescriptor) descriptors.get(j);
            	//if ( id.equals(existingDescriptor) /*&& id.getAssociatedSpecifications().contains(skelTree)*/  )
            	if ( id.equals(existingDescriptor) && id.getAssociatedSearchDescriptors().contains(skelTree)  )
            	//TODO Volvemos provisionalmente al concepto "fuerte" de igualdad de ndices porque el "dbil" da problemas.
            		//Por ejemplo:
            		// ndice usado para ? -> alpha . , ? , ? , 0
            		// y para ? -> alpha . ? beta , ? , ? , 0.
            		//Solamente indexamos por el cero, asi que con la linea comentada se consideraba el mismo.
            		//Sin embargo, un paso que espera lo primero cascara si obtiene lo segundo del ItemHandler.
            	{
            		System.out.println("Skeltree to add " + skelTree);
            		exists = true;
            		existingDescriptor.addSearchDescriptor(skelTree);
            		break;
            	}
            }
            
            if ( !exists )
            {
            	id.addSearchDescriptor(skelTree);
            	System.out.println("Adding index descriptor " + id);
            	descriptors.add(id);
            }
            
        }
        
        //ordenados por nmero de ndices atmicos. (i.e. los ndices ms
        //completos primero)
        Collections.sort(descriptors);
        
        descriptors.add(defaultIndex);
        
        return descriptors;
        
    }
    
    /**
     * Agrega todos los ndices atmicos (ndices sobre un nodo determinado y con una
     * estructura determinada) que pueda al descriptor de ndice dado, haciendo un recorrido
     * que parte del nodo dado de direccin dada.
     * @param id Descriptor de ndice al que queremos agregar ndices atmicos.
     * @param curAddress
     * @param curNode
     */
    private void addAllPossibleIndexes ( IndexDescriptor id , List curAddress , SearchDescriptorNode curNode )
    {
        Class curClass = curNode.getAssociatedClass();
        if ( curClass == null ) ; //no possible index for null
        else if ( curClass.isArray() ) //Object[], composite node.
        {
            for ( int i = 0 ; i < curNode.numChildren() ; i++ )
            {
                curAddress.add(new Integer(i));
                addAllPossibleIndexes(id,curAddress,curNode.getChild(i));
                curAddress.remove(curAddress.size()-1); //remove last added
            }
        }
        else  //Element, simple node.
        {
            

                	
            if ( element.HashIndexable.class.isAssignableFrom(curClass) )
            {
                List curAddressClone = new ArrayList();
                for ( int i = 0 ; i < curAddress.size() ; i++ )
                {
                    curAddressClone.add(curAddress.get(i));
                }
                id.addAtomicIndex( new AtomicIndex( curAddressClone , AtomicIndex.HASH_INDEX ) );
            }
            else       
            if ( element.ArrayIndexable.class.isAssignableFrom(curClass) )
            {
                //podemos crear un ndice array, lo hacemos.
                //como curAddress lo estamos modificando continuamente, utilizamos un clon. 
                List curAddressClone = new ArrayList();
                for ( int i = 0 ; i < curAddress.size() ; i++ )
                {
                    curAddressClone.add(curAddress.get(i));
                }
                id.addAtomicIndex( new AtomicIndex( curAddressClone , AtomicIndex.ARRAY_INDEX ) );
            }

        }
    }
    
    /**
     * Genera el cdigo de la declaracin de un atributo de la clase ItemHandler a partir
     * del descriptor de ndice que indica una estructura de ndice determinada. El cdigo
     * devuelto ser una declaracin de atributo que d soporte a dicha estructura de ndice
     * (por ejemplo, si es un ndice array sobre dos nodos del item, se devolver la
     * declaracin de un array bidimensional cuyo nombre ser el del IndexDescriptor).
     * @param id Descriptor del ndice que queremos implementar en un atributo.
     * @return Declaracin del atributo correspondiente a ese ndice para la clase ItemHandler.
     * @see IndexDescriptor
     */
    private String getAttributeDeclaration ( IndexDescriptor id )
    {
        
        List indices = id.getAtomicIndexes();
        
        StringBuffer decltype = new StringBuffer("");
        
        //este flag indica que an no se ha encontrado un tipo que nos impida continuar
        //aadiendo especificadores de tipo.
        //i.e. los arrays se pueden concatenar ([][]...) pero las HashMaps no, no hay
        //un tipo "tabla hash de tabla hash" al no haber genricos.
        boolean noBaseType = true;
        
        for ( int i = 0 ; i < indices.size() ; i++ )
        {
            AtomicIndex ai = (AtomicIndex) indices.get(i);
            
            if ( ai.getIndexType() == AtomicIndex.ARRAY_INDEX )
            {
                if ( noBaseType )
                    decltype.append("[]");
                else
                    decltype.append("/*of array*/ ");
            }
            else if ( ai.getIndexType() == AtomicIndex.HASH_INDEX )
            {
                if ( noBaseType )
                {
                    noBaseType = false;
                	decltype.insert(0,"HashMap");
                }
                else
                    decltype.append("/*of HashMap*/");
            }
        }
        
        if ( noBaseType )
            decltype.insert(0,"List /*of Object[]*/ ");
        else
            decltype.append("/*of List of Object[]*/ ");
        
        //decltype tiene el valor con el que hemos de declarar el ndice asociado a id.
        
        return decltype.toString() + " " + id.getName() + ";";
        
    }
    
    /**
     * Como getAttributeDeclaration; pero lo que genera es una inicializacin.
     */
    private String getAttributeInitialization ( IndexDescriptor id )
    {
        
        List indices = id.getAtomicIndexes();
        
        //llamada a constructor que va despus del new.
        StringBuffer constructor = new StringBuffer("");
        
        //este flag indica que an no se ha encontrado un tipo que nos impida continuar
        //aadiendo especificadores de tipo.
        //i.e. los arrays se pueden concatenar ([][]...) pero las HashMaps no, no hay
        //un tipo "tabla hash de tabla hash" al no haber genricos.
        boolean noBaseType = true;
        
        for ( int i = 0 ; i < indices.size() ; i++ )
        {
            AtomicIndex ai = (AtomicIndex) indices.get(i);
            
            if ( ai.getIndexType() == AtomicIndex.ARRAY_INDEX )
            {
                if ( noBaseType )
                {
                    String lenSpec = ai.getIndexedNodeClass().getName() + ".getArrayLength()";
                    constructor.append("["+lenSpec+"]");
                }
            }
            else if ( ai.getIndexType() == AtomicIndex.HASH_INDEX )
            {
                if ( noBaseType )
                {
                    noBaseType = false;
                	constructor.insert(0,"HashMap");
                }
            }
        }
        
        if ( noBaseType )
            constructor.insert(0,"ArrayList /*of Object[]*/ ");

        if ( !constructor.toString().endsWith("]") ) constructor.append("()");
        
        return id.getName() + " = new " + constructor.toString() + ";";
        
    }
    
    /**
     * Genera las declaraciones de atributos para los ndices dados.
     * @param indices Array de descriptores de ndice.
     * @return Cdigo para las declaraciones de los atributos que implementarn
     * los ndices dados.
     */
    protected String generateDeclarations ( IndexDescriptor[] indices )
    {
        StringBuffer result = new StringBuffer();
        for ( int i = 0 ; i < indices.length ; i++ )
        {
        	String comment = indices[i].toString();
            String s = getAttributeDeclaration(indices[i]);
            
            result.append("//");
            result.append(comment);
            result.append("\n");
            result.append(s);
            result.append("\n\n");
            
        }
        return result.toString();
    }
    
    
    protected String generateInitializations ( IndexDescriptor[] indices )
    {
        StringBuffer result = new StringBuffer();
        for ( int i = 0 ; i < indices.length ; i++ )
        {
            String s = getAttributeInitialization(indices[i]);
            result.append(s);
            result.append("\n");
        }
        return result.toString();
    }
    
    /**
     * <p>Genera el cdigo correspondiente a la obtencin de una lista de items (llamada
     * result, y que se supone ya declarada) a partir de un ndice dado, si se puede.</p>
     * <p>Tambin debe estar declarada una variable skelmatch.</p>
     * <p>La idea general del modo en que genera el cdigo es:
     * <ul>
     * <li>Para cada uno de los formatos de bsqueda asociados a ese ndice (i.e.
     * con los que se pueden usar ese ndice)</li>
     * <ul>
     * <li>Generar, mediante SearchDescriptorMatchCodeGenerator, un cdigo que se encarga
     * de determinar si el objeto pasado como parmetro al mtodo que estamos
     * generando corresponde a ese formato de bsqueda (descrito en un descriptor de bsqueda),</li>
     * <li>Generar el cdigo que, en caso afirmativo, devuelve una lista de items
     * obtenida de este ndice. Ntese que utilizamos los resultados parciales
     * que nos devuelve el IndexDescriptor para introducir variables parciales
     * con las que hacer comprobaciones de valor no nulo.</li>
     * </ul>
     * </ul>
     * </p>
     * @param index ndice del que se obtendrn los items.
     * @param objectName Nombre de la variable que contiene el tem que sirve como
     * especificacin o peticin.
     * @param generateResultCleanupCode: si se chequear que las partes de los items
     * de la lista resultado que no correspondan a elementos por los que indexamos
     * se ajustan al criterio pedido en la especificacin.
     * @return El cdigo generado. Se supone declarada la variable result, que termina
     * a null si no se encuentra ningn item y a una lista de items de lo contrario.
     */
    protected String generateSinglePetitionCode ( IndexDescriptor index , String objectName , boolean expectGroundStrings , boolean generateResultCleanupCode , boolean generateResultExistenceCheckCode , boolean nullsAllowedAtInput )
    {
        
        StringBuffer result = new StringBuffer("");
        
        List skeletonTrees = index.getAssociatedSearchDescriptors();
        
        //poner un comentario con la descripcin del ndice.
        result.append("//" + index.toString()+"\n");
        
        for ( int i = 0 ; i < skeletonTrees.size() ; i++ )
        {
            SearchDescriptorNode currentSkeletonTree = (SearchDescriptorNode) skeletonTrees.get(i);
            
            SearchDescriptorMatchCodeGenerator smcg = new SearchDescriptorMatchCodeGenerator(objectName);
            
            result.append("skelmatch = true;\n");
            
            result.append( (smcg.generateCode(currentSkeletonTree,expectGroundStrings,nullsAllowedAtInput )) );
            
            result.append("if ( skelmatch )\n");
            result.append("{ //matched skeleton" + i +"\n");
            
            String usageCode = index.getIndexUsageCode(smcg.getAddressConverter(),false);
            List partialResults = index.getPartialIndexUsageCodeResults();
            
            
            if ( nullsAllowedAtInput )
            {
            	result.append("try\n");
            	result.append("{\n");
            }
            
            
            for ( int j = 0 ; j < partialResults.size() ; j++ )
            {
            	IndexDescriptor.PartialAccessResult partialResult = (IndexDescriptor.PartialAccessResult) partialResults.get(j);
            	result.append ( partialResult.getDeclarationCode() + " partial" + j + ";\n");
            	result.append ( "if ( (partial"+j+"="+partialResult.getAccessCode()+")==null)\n");
            	result.append ( "result = null;\n");
            	//if ( j < partialResults.size() - 1 )
            	//{
            		result.append ( "else\n");
            	/*unc*/	result.append ( "{\n" );
            	//}            	
            	//usar la variable parcial en el cdigo subsiguiente
            	usageCode = usageCode.replaceFirst("\\Q"+partialResult.getAccessCode()+"\\E","partial"+j);
            	for ( int k = j+1 ; k < partialResults.size() ; k++ )
            	{
            		((IndexDescriptor.PartialAccessResult)partialResults.get(k)).replaceAccessCode("\\Q"+partialResult.getAccessCode()+"\\E","partial"+j);
            	}
            }
                        
            result.append("result = ");
                       
            result.append( usageCode );
                   
            result.append(";\n");
                                 
            if ( generateResultCleanupCode )
            {
            
	            String resultCleanupCode = smcg.generateResultCleanupCode ( index , currentSkeletonTree , false );
	            
	            if ( resultCleanupCode != null )
	            	result.append ( resultCleanupCode );
            
            }
            
            if ( generateResultExistenceCheckCode )
            {
	            String resultExistenceCheckCode = smcg.generateResultCleanupCode ( index , currentSkeletonTree , true );
	            
	            if ( resultExistenceCheckCode != null )
	            	result.append ( resultExistenceCheckCode );
            }
                      
            //cerrar todas las llaves de los ifs del bucle anterior menos la primera
            for ( int j = 0 ; j < partialResults.size()  ; j++ )
            {
                result.append("}\n");
            }
            
            if ( nullsAllowedAtInput )
            {
            	result.append("} //try block\n");
            	result.append("catch ( NullPointerException npe )\n");
            	result.append("{\n");
            	result.append("result = new ArrayList();\n");
            	result.append("}\n");
            	result.append("finally\n");
            	result.append("{\n");
            	result.append("if ( result == null ) result = "+ index.getName() + "_alt;\n");
            	
            	//05-04-14 bugfix: clone() needed in order not to alter internal structure
            	result.append("else\n");
            	result.append("{\n");
            	result.append("result = (List)((ArrayList)result).clone();\n");
            	result.append("result.addAll("+index.getName() + "_alt)" + ";\n");
            	result.append("}\n");
            	
				//pre 05-04-14 code line
				//result.append( "else result.addAll("+index.getName() + "_alt)" + ";\n" );
            	
            	result.append("}\n");
            }
                     
            result.append("} //end if matched search descriptor " + i + "\n");

            if ( i < skeletonTrees.size() - 1 )
            {
                result.append("else\n");
                result.append("{ //didn't match search descriptor " + i + "\n");           
            }
                
        }
        
        //cerrar llaves abiertas, que son skeltrees.size() - 1
        for ( int i = 0 ; i < skeletonTrees.size()-1  ; i++ )
        {
            result.append("}\n");
        }
        
        if ( skeletonTrees.size() == 0 )
        {
            //caso especial sin ningn descriptor de bsqueda asociado: es siempre el ndice vaco.
            result.append("skelmatch = true;\n");
            result.append("result = ");
            result.append( index.getIndexUsageCode(new HashMap(),false));
            result.append(";\n");
        }
        
        
        return result.toString();
        
    }
    
    /**
     * Genera el cdigo de aadir un item de nombre objectName al ndice dado, si procede.
     * (es decir, si el item tiene un formato compatible con alguno de los descriptores de bsqueda
     * asociados a ese ndice)
     * @param index ndice que estamos considerando.
     * @param objectName Nombre de la variable que contiene el item.
     * @return Cdigo que agrega el item al ndice si forma parte del conjunto de items
     * que dicho ndice puede indexar.
     */
    protected String generateSingleAddCode ( IndexDescriptor index , String objectName )
    {
        
        StringBuffer result = new StringBuffer("");
        
        List skeletonTrees = index.getAssociatedSearchDescriptors();
        
        
        for ( int i = 0 ; i < skeletonTrees.size() ; i++ )
        {
            SearchDescriptorNode currentSkeletonTree = (SearchDescriptorNode) skeletonTrees.get(i);
            
            SearchDescriptorMatchCodeGenerator smcg = new SearchDescriptorMatchCodeGenerator(objectName);
            
            result.append("skelmatch = true;\n");
            
            result.append( (smcg.generateCode(currentSkeletonTree,true,false)) );
            
            result.append("if ( skelmatch )\n");
            result.append("{\n");
            
            String indexUsageCode = index.getIndexUsageCode(smcg.getAddressConverter(),false);
            String initterCode = index.getIndexUsageCode(smcg.getAddressConverter(),true);
            
            List partialResults = index.getPartialIndexUsageCodeResults();
            
            for ( int j = 0 ; j < partialResults.size() ; j++ )
            {
            	IndexDescriptor.PartialAccessResult partialResult = (IndexDescriptor.PartialAccessResult) partialResults.get(j);
            	result.append ( partialResult.getDeclarationCode() + " partial" + j + ";\n");
            	result.append ( "if ( (partial" + j + "=" + partialResult.getAccessCode() + " ) == null )\n");
            	result.append ("{\n");
            	result.append ( partialResult.getInitializerCode() + ";\n");
            	result.append ( "partial"+j + " = " + partialResult.getAccessCode() + ";\n" );
            	result.append ("}\n");
            	//usar la variable parcial en el cdigo subsiguiente
            	indexUsageCode = indexUsageCode.replaceFirst("\\Q"+partialResult.getAccessCode()+"\\E","partial"+j);
            	for ( int k = j+1 ; k < partialResults.size() ; k++ )
            	{
            		((IndexDescriptor.PartialAccessResult)partialResults.get(k)).replaceAccessCode("\\Q"+partialResult.getAccessCode()+"\\E","partial"+j);
            	}
            }
            
            
            result.append("if ( " + indexUsageCode + " == null )\n");
            
            
            result.append(initterCode);
            
            result.append("(");
            result.append( indexUsageCode );
            result.append(").add("+objectName+");\n");
            result.append("}\n");
            if ( i < skeletonTrees.size() - 1 )
            {
                result.append("else\n");
                result.append("{\n");           
            }
        }
        
        //cerrar llaves abiertas, que son skeletonTrees.size() - 1 debido al ltimo if
        //del bucle anterior
        for ( int i = 0 ; i < skeletonTrees.size() - 1 ; i++ )
        {
            result.append("}\n");
        }
        
        if ( skeletonTrees.size() == 0 )
        {
            //caso especial sin ningn esqueleto asociado: es siempre el ndice vaco.
            //(lista de items)
            String indexUsageCode = index.getIndexUsageCode(new HashMap(),false);
            result.append("(");
            result.append( indexUsageCode );
            result.append(").add("+objectName+");\n");            
        }
        
        return result.toString();
        
    }

    private String generategetBySpecificationMethod ( IndexDescriptor[] indices , String objectName )
    {
        StringBuffer result = new StringBuffer("public List getBySpecification( Object[] " + objectName + ")\n");
        result.append("{\n");
        result.append("List result = null;\n");
        result.append("boolean skelmatch;\n");
        for ( int i = 0 ; i < indices.length ; i++ )
        {
            result.append(generateSinglePetitionCode(indices[i],objectName,false,true,false,false));
            result.append("if ( skelmatch ) return result;\n");
        }
        result.append("return null;\n");
        result.append("}//[getBySpecification]\n");
        return result.toString();
    }
    
    //este mtodo no est todo lo optimizado que podra estar; pero tampoco es un cuello de botella.
    private String generateExistsMethod ( IndexDescriptor[] indices , String objectName )
    {
        StringBuffer result = new StringBuffer("public boolean exists( Object[] " + objectName + ")\n");
        result.append("{\n");
        result.append("List result = null;\n");
        result.append("boolean skelmatch;\n");
        for ( int i = 0 ; i < indices.length ; i++ )
        {
            result.append(generateSinglePetitionCode(indices[i],objectName,true,true,true,false));
            
            
            
            
            //can't use contains for arrays, equals is defined as ==:
            result.append("if ( skelmatch )\n");
            result.append("{\n");
            result.append("if ( result == null ) return false;\n");
            result.append("int sz = result.size();\n");
            result.append("for ( int cind = 0 ; cind < sz ; cind++ )\n");
            result.append("{\n");
            result.append("Object[] itemToCompare = ((Object[])result.get(cind));\n");            
            result.append("if ( compareItems(item,itemToCompare) )\n");
            result.append("{\n");
            result.append("return true;\n");
            result.append("}\n");
            
            result.append("}//[result list for]\n");
            result.append("return false;\n");
            result.append("}//[if skelmatch]\n");
        }
        result.append("return false;\n");
        result.append("}//[exists]\n");
        return result.toString();
    }
    
    private String generateAddItemMethod ( IndexDescriptor[] indices , String objectName )
    {
        StringBuffer result = new StringBuffer("public void addItem( Object[] " + objectName + ")\n");
        result.append("{\n");
        result.append("boolean skelmatch;\n");
        for ( int i = 0 ; i < indices.length ; i++ )
        {
            result.append(generateSingleAddCode(indices[i],objectName));
        }
        result.append("}//[addItem]\n");
        return result.toString();
    }
    
    public String generateCode ( SearchDescriptorNode[] searchSkeletonStructures , SearchDescriptorNode[] existenceSkeletonStructures )
    {
        
        StringBuffer codeBuffer = new StringBuffer("");
        
        codeBuffer.append("package schema.item;\n\n");
        codeBuffer.append("import java.util.*;\n");
        
        codeBuffer.append("\n");
		codeBuffer.append("public class ItemHandler\n");
		codeBuffer.append("{\n\n");
		codeBuffer.append("//Attributes:\n");
        
        //obtenemos los descriptores de los ndices que deberemos crear.
        List /*of IndexDescriptor*/ indices = getSuitableIndices ( "index" , "itemList" , searchSkeletonStructures );
        List /*of IndexDescriptor*/ existenceIndices = getSuitableIndices ( "ex_index" , "itemList" , existenceSkeletonStructures );
        
        if ( !EXISTENCE_INDEXING )
        	existenceIndices = new ArrayList();
        
        List allIndices = new ArrayList(); //Lista de ndices sin repeticiones.
        
        //quitamos ndices duplicados de la lista de ndices de bsqueda.
        //ntese que el ndice por defecto no se quitar, porque est de ltimo
        //y en caso de duplicado siempre se quita el primero.
        for ( int i = 0 ; i < indices.size() ; i++ )
        {
        	for ( int j = i+1 ; j < indices.size() ; j++ )
        	{
        		if ( indices.get(i).equals(indices.get(j)) )
        		{
        			//TODO Esto no es compatible con la definicin fuerte de igualdad de ndices.
        			//De esta manera las cosas van mal si hay ndices iguales (ej [0][1][2][3][4]) sobre esqueletos distintos,
        			//porque SLO NOS QUEDAMOS CON UN ESQUELETO.
        			indices.remove(i);
        			i--; //porque haremos i++ luego, y queremos quedarnos fijos
        			break;
        		}
        	}
        }
        
        //hacemos lo mismo con la lista de ndices de existencia.
        for ( int i = 0 ; i < existenceIndices.size() ; i++ )
        {
        	for ( int j = i+1 ; j < existenceIndices.size() ; j++ )
        	{
        		if ( existenceIndices.get(i).equals(existenceIndices.get(j)) )
        		{
        			existenceIndices.remove(i);
        			i--; //porque haremos i++ luego, y queremos quedarnos fijos
        			break;
        		}
        	}
        }
        
        //quitamos los ndices de existencia que sean iguales a ndices de bsqueda,
        //reemplazando sus referencias por referencias a los de bsqueda.
        //los ndices que sean exclusivamente de existencia los aadimos a allIndices.
        boolean exists = false;
        for ( int i = 0 ; i < indices.size() ; i++ )
        {
        	if ( existenceIndices.contains(indices.get(i)) )
        	{
        		existenceIndices.remove(existenceIndices.indexOf(indices.get(i))); //ndice igual pero con otro nombre.
        		existenceIndices.add(indices.get(i)); //el mismo ndice, con el mismo nombre.		
        	}
        	else
        	{
        		allIndices.add(indices.get(i));
        	}
        }   
        for ( int i = 0 ; i < existenceIndices.size() ; i++ )
        {
        	allIndices.add(existenceIndices.get(i));
        }
        
        	
        
        IndexDescriptor[] allIndicesAr = (IndexDescriptor[]) allIndices.toArray(new IndexDescriptor[0]);  
        IndexDescriptor[] searchIndicesAr = (IndexDescriptor[]) indices.toArray(new IndexDescriptor[0]);  
        IndexDescriptor[] existenceIndicesAr = (IndexDescriptor[]) existenceIndices.toArray(new IndexDescriptor[0]);  
        
  
        if ( !EXISTENCE_INDEXING )
        	existenceIndicesAr = allIndicesAr;
        
        codeBuffer.append(generateDeclarations(allIndicesAr));
        
        codeBuffer.append("private static ItemHandler instance;\n");
        codeBuffer.append("private static List savedItemList; //permite guardar la lista de items\n"); 
        codeBuffer.append("\n\n");
        
        codeBuffer.append("//Methods:\n");
        
		try
		{
			//item compare method: read from file.
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream("common"+File.separator+"parts"+File.separator+"ItemHandler.p1") ) );
			String linea;
			while ( (linea=br.readLine()) != null )
			{
				codeBuffer.append(linea);
				codeBuffer.append("\n");
			}
		}
		catch ( Exception ioe )
		{
			System.out.println("Problem when reading partial code file common"+File.separator+"parts"+File.separator+"ItemHandler.p1:");
			ioe.printStackTrace();
		}
        
        codeBuffer.append("public ItemHandler()\n");
        codeBuffer.append("{\n");
        codeBuffer.append(generateInitializations(allIndicesAr));        
        codeBuffer.append("instance = this;\n");
        codeBuffer.append("}//[constructor]\n");
        
        codeBuffer.append("\n");

        codeBuffer.append("public static ItemHandler instance()\n");
        codeBuffer.append("{\n");
        codeBuffer.append("return instance;\n");
        codeBuffer.append("}//[singleton access method]\n");
               
        codeBuffer.append("\n");
        
        codeBuffer.append("public List getAll()\n");
        codeBuffer.append("{\n");
        codeBuffer.append("return itemList;\n");
        codeBuffer.append("}//[getAll]\n");
        
        codeBuffer.append( this.generategetBySpecificationMethod(searchIndicesAr,"item") );
        
        codeBuffer.append("\n");
        
        codeBuffer.append( this.generateExistsMethod(existenceIndicesAr,"item") );
        
        codeBuffer.append("\n");
        
        codeBuffer.append( this.generateAddItemMethod(allIndicesAr,"item") );
        
        codeBuffer.append("\n\n} //[class]\n\n");
        
        return codeBuffer.toString();
        
    }
    
    


}


