package grammar.tag.util;

import java.io.File;
import java.io.IOException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import util.DOMNodeListWrapper;
import util.DOMNodeWrapper;


/*
 * Applies a transformation to a grammar in the TAGML format.
 * 
 * Methods preprocessTreeNode(), postprocessTreeNode(),
 * preprocessNodeNode() and postprocessNodeNode() are empty
 * template methods to be overridden with subclasses.
 * 
 * This transformer also supports the chain of responsibility
 * design pattern in order to sequentially chain several transformers.
 */
public abstract class AbstractGrammarTransformer 
{

	
	private String inFile;
	private String outFile;
	private AbstractGrammarTransformer next;
	
	public AbstractGrammarTransformer ( String inFile , String outFile )
	{
		this.inFile = inFile;
		this.outFile = outFile;
		this.next = null;
	}
	
	public AbstractGrammarTransformer ( String inFile , String outFile , AbstractGrammarTransformer next )
	{
		this.inFile = inFile;
		this.outFile = outFile;
		this.next = next;
	}
	
	public AbstractGrammarTransformer ( AbstractGrammarTransformer next )
	{
		this.inFile = null;
		this.outFile = null;
		this.next = next;
	}
	
	public boolean hasNext()
	{
		return ( next != null );
	}
	
	public AbstractGrammarTransformer next()
	{
		return next;
	}
	
	public String toString()
	{
		return "Abstract Grammar Transformer";
	}
	
	public void traverseTreeNode ( DOMNodeWrapper treeNode )
	{
		preprocessTreeNode(treeNode);
		DOMNodeListWrapper rootNodes = treeNode.getDirectChildrenElementsByTagName("node");
		for ( int i = 0 ; i < rootNodes.size() ; i++ )
		{
			traverseNodeNode ( rootNodes.get(i) );
		}
		postprocessTreeNode(treeNode);
	}
	
	public void traverseNodeNode ( DOMNodeWrapper nodeNode )
	{
		preprocessNodeNode(nodeNode);
		DOMNodeListWrapper childNodes = nodeNode.getDirectChildrenElementsByTagName("node");
		for ( int i = 0 ; i < childNodes.size() ; i++ )
		{
			traverseNodeNode ( childNodes.get(i) );
		}
		postprocessNodeNode(nodeNode);
	}
	
	public void preprocessTreeNode(DOMNodeWrapper treeNode){;}
	public void postprocessTreeNode(DOMNodeWrapper treeNode){;}
	public void preprocessNodeNode(DOMNodeWrapper nodeNode){;}
	public void postprocessNodeNode(DOMNodeWrapper nodeNode){;}
	
	
	public void transform ( Document document )
	{
		DOMNodeWrapper root = new DOMNodeWrapper ( document.getDocumentElement() );
		
		DOMNodeWrapper rootTrees = root;
		
		System.out.println(toString());
		
		//tree families
		DOMNodeListWrapper famNodes = rootTrees.getElementsByTagName("family");
		for ( int i = 0 ; i < famNodes.size() ; i++ )
		{
			DOMNodeListWrapper treeNodes = famNodes.get(i).getElementsByTagName("tree");
			for ( int j = 0 ; j < treeNodes.size() ; j++ )
			{
				DOMNodeWrapper treeNode = treeNodes.get(j);
				traverseTreeNode(treeNode);
			}
		}
		
		//single trees
		DOMNodeListWrapper singNodes = rootTrees.getElementsByTagName("single-tree");
		for ( int i = 0 ; i < singNodes.size() ; i++ )
		{
			DOMNodeListWrapper treeNodes = singNodes.get(i).getElementsByTagName("tree");
			for ( int j = 0 ; j < treeNodes.size() ; j++ )
			{
				DOMNodeWrapper treeNode = treeNodes.get(j);
				traverseTreeNode(treeNode);
			}
		}
		
		if ( hasNext() )
			next.transform(document);
		
	}
	
	public void process () throws ParserConfigurationException , IOException, SAXException, TransformerFactoryConfigurationError, TransformerException
	{
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setValidating(true);
		factory.setNamespaceAware(true);
		DocumentBuilder builder = factory.newDocumentBuilder();
		Document document = builder.parse( new File(inFile) );
		
		this.transform(document);
		
		Transformer t = TransformerFactory.newInstance().newTransformer();
		t.setOutputProperty(OutputKeys.INDENT, "yes");
		t.transform(new DOMSource(document),new StreamResult(new File(outFile)));
	}
	
	
	
}
